import java.util.*;
import java.io.*;
import AST.*;
import CompilerError.*;
import Interface.*;


/**
  In directory basiclib should be put arrays of all basic types up to dimension
  three ( array(byte)[], array(byte)[][], array(byte)[][][], ... ), the basic 
  classes (char, byte, integer, ...), all classes used by the 
  array classes (Iter$Array$1$byte, ...).
      In directory exceptions should be put all exception classes. 
      In directory greenlib should be  put the basic green library (Vector, Filter, ...)
      
  Grammar for the project
    Project ::= FileNameList  OPTIONS OptionsList 
    FileNameList ::= | FileNameList FileName
    OptionsList ::= Option | OptionsList Option
    Option ::= BASICLIB "=" FileName | LIB "=" FileName | WRAPPER "=" FileName 
              | USEIRL | COMPILEONLY | COMPILEBASICCLASSES 
              | ARRAYDIR "=" FileName | CLASSDIR "=" FileName
              | JAVALIB "=" FileName
              | TEMPLATEDIR  "=" FileName
              | PROJECTDIR = FileName
              | SUBTYPEEQUALSUBCLASS
              | INTERFACEDIR = FileName
    FileName ::= STRING | PATHNAME
    
*/


public class GreenProject {
    
    public boolean init( String fileName, ProjectError error,  ErrorMsg errorMsg  ) {
        this.errorMsg = errorMsg;
        this.error = error;
        MyFile myFile = new MyFile(fileName);
        input = myFile.readFile();
        if ( myFile.getError() != MyFile.ok_e ) 
           return false;
        setDefaultOptions();
        arrayPathLib = new Vector();
        arrayJavaLib = new Vector();
        dirList = new Vector();
        //c classList = new ClassList();
        toBeCompiledList = new Vector();
        toCreateClassList = new ToCreateClassList();
        toCreateArrayList = new ToCreateArrayList();
        separator = File.separator;
        projectDir = "";
        return true;
    }
    
    private void setDefaultOptions() {
        subtypeEqualSubclass = false;
        useIRL = false;
        compileOnly = false;
        compileBasicClasses = false;
        arrayDir = "arraytmp/";
        classDir = "classtmp/";
        interfaceDir = "interface/";
        templateDir = null;
    }

    public void firstTime() {
       createAuxArrayClasses(); 
    }

    
    private boolean compileClass( ClassFile cf, ClassTypeInterface cti ) {
        
       String fileName = cf.getClassName() + ".g";
       
       String pathFile = cf.getDirectory() + separator + fileName;
       MyFile myFile = new MyFile(pathFile);
       input = myFile.readFile();
       if ( myFile.getError() != MyFile.ok_e )  {
          error.show("Can't read file " + pathFile);
          return false;
       }
       else {
          MyLexer lexer;
          try {
            lexer = new MyLexer( input, errorMsg );
          } catch ( Exception e ) {
            error.show("Error reading file " + pathFile);
            return false;
          }
          errorMsg.setLexer(lexer, pathFile);
          
          MyParser parser;

          try { 
             parser = new MyParser(lexer, errorMsg, cti, til);
          } catch ( Throwable e ) {
            return false;
          }

          GenericClass aClass;
          try {
            aClass = (GenericClass ) parser.parse();
          } catch ( Exception e ) {
            System.out.println("Just before printStackTrace");
            e.printStackTrace();
            //error.show("Error compiling " + pathFile);
            return false;
          }
          
          if ( aClass == null ) 
             return false;

          String className;
          if ( TypeInterface.basicTypes.get(cti.getName()) != null ||
               cti.getName().compareTo("String") == 0 )  
             className = TypeInterface.buildBasicTypeClassName(cti.getName());
          else 
             className = cti.getJavaName();
         

          String outClassName = cf.getDirectory() + separator + className + ".java";
          String outClassObjectName = cf.getDirectory() + separator + 
                 cti.getClassObjectInterface().getInterfaceName() + ".java";

          
          try {
             PrintWriter outClass = new PrintWriter( 
                 new FileOutputStream(outClassName) );
                 
             //if ( basicClassesTable.get(aClass.getName()) != null ) 
                  // if not a basic class, compile it.
             aClass.gen(outClass, useIRL, className);
             aClass.genInterfaces(projectDir + separator + interfaceDir, errorMsg);
             //}
             if ( cti.getIsShellClass() ) {
                    String fileName_ivc = cf.getDirectory() + separator + 
                    aClass.getShell_ivcName() + ".java";
                try {
                   PrintWriter out_ivc = new PrintWriter( 
                         new FileOutputStream(fileName_ivc) );
                   aClass.genShell_ivc(out_ivc);
                   out_ivc.close();
                }
                catch ( Exception excep ) {
                   errorMsg.signal("Error writing " + fileName_ivc );
                }
             }
             else {
                PrintWriter outClassObject = new PrintWriter( 
                   new FileOutputStream(outClassObjectName) );
                 
                
                ClassObject classObject = aClass.getClassObject();
                classObject.gen(outClassObject, useIRL);
                classObject.genInterfaces(projectDir + separator + interfaceDir, errorMsg);
                outClassObject.close();
             }
             
             outClass.close();
          }
          catch( Exception excep ) {
             excep.printStackTrace();
             error.show("Error writing to file " + outClassName + " or " + outClassObjectName);
             return false;
          }
          return true;
       }
    }



    
    public void make() {

        ClassFile cf;
        
          // number of the current line
        lineNumber = 1;
        tokenPos = 0;
        beforeOptions = true;
        inFileName = true;
        cfl = new Vector();
        ccl = new Vector();
        til = new TypeInterfaceList();
        pcl = new Hashtable();
        
        nextToken();
        parse();
        Enumeration e;
        if ( error.getWasACompilerError() ) 
           return;
           
            
        boolean result = addArrayLib();
        if ( ! addBasicLib() )
           result = false;
        if ( ! addWrapperLib() ) 
           result = false;
        
        
        e = cfl.elements();
        while ( e.hasMoreElements() ) {
           cf = (ClassFile ) e.nextElement();
           String directory = cf.getDirectory();
           String name = cf.getClassName();
                
           File fg = new File(directory + separator + name + ".g");
           File fgi = new File(directory + separator + name + ".gi");
           File fj = new File(directory + separator + 
              TypeInterface.buildRealClassJavaName(name) + ".java");
           
              
           if ( ! fg.exists() ) {
                 // .g file does not exists
               if ( ! fgi.exists() ) {
                     // .gi files does not exist
                   error.show("File " + fgi.getPath() + " cannot be read", -1);
               }
               else 
                     // .gi file exists
                  readInterface(cf, true);
           }
           else 
                 // fg exists
               if ( ! fgi.exists() || ! fj.exists() || 
                    fg.lastModified() > fgi.lastModified() ) {
                   // Either .gi or .java does not exists or .gi is older than ".g". 
                   // Therefore, ".g" needs to be recompiled
                   if ( readInterface(cf, false) ) {
                      ccl.addElement(cf);
                   }
               }
               else {
                   // .gi exists and .gi is newer than .g
                   readInterface(cf, true);
               }
        }
         
        if ( error.getWasACompilerError() ) 
           return;
        // create instantiations of parameterized classes and array classes
        createClassesFromClassList();
        if ( error.getWasACompilerError() ) 
           return;
        
        errorMsg.setShowNameLineColumn(false);
           // reify all interfaces
        reifyBasicClasses();
        if ( error.getWasACompilerError() ) 
           return;
        e = til.elements();
        while ( e.hasMoreElements() ) {
            String interfaceName = (String ) e.nextElement();
           ((ClassTypeInterface ) til.get(interfaceName)).reify(til, errorMsg);
        }
        if ( error.getWasACompilerError() ) 
           return;
        errorMsg.setShowNameLineColumn(true);
        e = ccl.elements();
        while ( e.hasMoreElements() ) {
           cf = (ClassFile ) e.nextElement();
           ClassTypeInterface cti = ((ClassTypeInterface ) til.get( cf.getClassName() ));
           if ( cti == null ) {
              errorMsg.show("Internal error: cti == null");
           }
           if ( ! cti.getIsShellClass() ) {
              String filename = cf.getDirectory() + separator + cf.getClassName() + ".gi";

              try {
                 PrintWriter outInterface = new PrintWriter( 
                    new FileOutputStream(filename) );
                 cti.genI(outInterface);
                 outInterface.close();
              }
              catch( Exception excep) {
                 error.show("Error writing to file " + filename);
              }
           }
        }
           
        createDirectories();
        
        e = ccl.elements();
        while ( e.hasMoreElements() ) {
           cf = (ClassFile ) e.nextElement();
//           if ( cf.getClassName().compareTo("A") == 0 ||
//                cf.getClassName().compareTo("B") == 0 ||
//                cf.getClassName().compareTo("C") == 0 )  {
           ClassTypeInterface cti = ((ClassTypeInterface ) til.get( cf.getClassName() ));
           compileClass(cf, cti);
//                }
        }
        prepareBatFile();
        System.out.println("This is the end !!!");
    }

   public boolean prepareBatFile() {

      dirList.addElement( projectDir );
      dirList.addElement( projectDir + separator + arrayDir );
      dirList.addElement( projectDir + separator + classDir );
      dirList.addElement( projectDir + separator + interfaceDir );
      dirList.addElement( pathBasicLib );
      dirList.addElement( pathWrapper );
      Enumeration e = arrayPathLib.elements();
      while ( e.hasMoreElements() ) 
         dirList.addElement( (String ) e.nextElement() );
          
      String outFileName = projectDir + separator + "do.bat";
      try {
         PrintWriter out = new PrintWriter( 
                   new FileOutputStream(outFileName) );
                   
         out.print("javac ");
         
         out.print("-sourcepath ");
         int size = dirList.size();
         e = dirList.elements();
         while ( e.hasMoreElements() )  {
            out.print( (String ) e.nextElement() );
            if ( --size > 0 ) 
               out.print(";");
         }
         /*
         String classPath = System.getProperty("java.class.path");
         if ( classPath != null && classPath.length() > 0 )
            out.print(classPath);
         */
         
         out.println(" @filelist.txt");
         out.close();    
         
         outFileName = projectDir + separator + "filelist.txt";
         
         out = new PrintWriter( new FileOutputStream(outFileName) );

      
         e = ccl.elements();
         while ( e.hasMoreElements() ) {
            ClassFile cf = (ClassFile ) e.nextElement();
            if ( cf.getClassName().compareTo("String") != 0 ) {
               ClassTypeInterface cti = cf.getClassTypeInterface();
               out.println( cf.getDirectory() + separator 
                  + cf.getJavaFileName() + ".java" );
               if (  cti.getIsShellClass() ) 
                  out.println( cf.getDirectory() + separator + 
                     cti.getShell_ivcName() + ".java" );
               else
                  out.println( cf.getDirectory() + separator + 
                     TypeInterface.buildClassObjectName(cf.getClassName()) 
                     + ".java" );
            }
         }
         
         String startClassObjectName = TypeInterface.buildClassObjectName(startClass);
         outFileName = projectDir + separator + "Main.java";
         out.println(outFileName);
         out.close();
         
         out = new PrintWriter( new FileOutputStream(outFileName) );
         ws.set(out);
         ws.println("public class Main {");
         ws.add();
         ws.println("public static void main( String args[] ) {");
         ws.add();
         ws.println(startClassObjectName + 
             " start = _" + startClass + ".$classObject;" );
         ws.println("_Runtime.$classObject._setCatchUnchecked$_CatchUncheckedException( _HCatchUncheckedException.$classObject._new() );");
         ws.println("initialize();");
         
         
         ClassObjectTypeInterface coti = til.get(startClass).getClassObjectInterface();
         // stopped here
         String runMethodName, runStringMethodName;
         ParameterList pl = new ParameterList();
         runMethodName = TypeInterface.buildMethodSearchName("run", pl ); 
         
         Parameter p = new Parameter("args"); 
         ArrayType arrayType = new ArrayType( Type.stringType, 1 );
         p.setType(arrayType);
         pl.add(p);
         runStringMethodName = TypeInterface.buildMethodSearchName("run", pl); 
         
         MethodInterface runMethod, runStringMethod;
         runMethod = coti.publicSearch(runMethodName);
         runStringMethod = coti.publicSearch(runStringMethodName);
         if ( runMethod != null && runStringMethod != null ) {
            error.show("Error: the class in which the execution starts, " + startClass + " should contain just only method run");
            return false;
         }
         if ( runMethod != null ) {
            ws.println("start._run();");
         }
         else
            ws.println("start._run(args);");
         ws.sub();
         ws.println("}");

         ws.println("public static void initialize() {");
         ws.add();  

         Vector fileList = new Vector();
         e = dirList.elements();
         while ( e.hasMoreElements() )  {
            File file = new File( (String ) e.nextElement() );
            String allFiles[] = file.list();
            for( int i = 0; i < allFiles.length; i++ ) 
               if (  allFiles[i].endsWith(".java") ) {
                  String className = allFiles[i].substring(0, allFiles[i].lastIndexOf('.'));
                  if ( ! className.startsWith("$co$") &&
                       ! className.startsWith("$mi$") &&
                       (className.charAt(0) == '_' ||
                        className.charAt(0) == '$')  ) {
                     ws.println(className + ".initialize();");
                     fileList.addElement(className);
                  } 
               }
         }

         ws.println("_Runtime.$classObject._$setClasses$$Array$1$_ClassInfo(");
         ws.add();
         ws.println("new $Array$1$_ClassInfo (");
         ws.add();
         ws.println("new _ClassInfo[] {");
         ws.add();
         
         size = fileList.size();
         e = fileList.elements();
         while ( e.hasMoreElements() ) {
            String className = (String ) e.nextElement();
            if ( className.indexOf("$shell") >= 0 ) 
              size--;
            else {
               ws.print( className + ".$i" );
               if ( --size > 0 )
                 out.println(",");
               else
                  out.println();
            }
         }
         ws.println("}");
         ws.sub();
         ws.println(")");
         ws.sub();
         ws.println(");");
         ws.sub();     
         
         ws.sub();
         ws.println("}");
         
         ws.sub();
         out.println("}");
         out.close();

         outFileName = projectDir + separator + "run.bat";
         out = new PrintWriter( 
                   new FileOutputStream(outFileName) );
         out.print("java -classpath ");
         size = dirList.size();
         e = dirList.elements();
         while ( e.hasMoreElements() )  {
            out.print( (String ) e.nextElement() );
            if ( --size > 0 ) 
               out.print(";");
         }
         out.println(" Main");
         out.close();
         
      }
      catch( Exception excep ) {
         System.out.println("Exception " + excep.toString());
         excep.printStackTrace();
         error.show("Error writing to file " + outFileName);
         return false;
      }
      return true;
   }
        
        
   public void createDirectories() {
    
      File arrayDirFile = new File(projectDir + separator + arrayDir);
      File classDirFile = new File(projectDir + separator + classDir);
      File interfaceDirFile = new File(projectDir + separator + interfaceDir);
      
      if ( ! arrayDirFile.exists() )
         arrayDirFile.mkdirs();
      if ( ! classDirFile.exists() )
         classDirFile.mkdirs();
      if ( ! interfaceDirFile.exists() )
         interfaceDirFile.mkdirs();
   }


   public boolean readInterface( ClassFile cf, boolean is_gi ) {
    
       String pathFile = cf.getDirectory() + separator + cf.getClassName();
       pathFile = pathFile + (is_gi ? ".gi" : ".g");
       
       MyFile myFile = new MyFile(pathFile);
       input = myFile.readFile();
       if ( myFile.getError() != MyFile.ok_e ) {
           error.show("Error reading file " + pathFile);
           return false;
       }
       else {
          MyLexer lexer;
          try {
            lexer = new MyLexer( input, errorMsg );
          } catch ( Exception e ) {
            error.show("Internal error in readInterface");
            return false;
          }
          errorMsg.setLexer(lexer, pathFile);
          
          InterfaceParser parser;
          boolean ok;

          
          try { 
             parser = new InterfaceParser( 
                   lexer, errorMsg, toCreateClassList, toCreateArrayList, is_gi ); 
                   
             ok = parser.parse();
          } catch ( Throwable e ) {
            e.printStackTrace();
            error.show("Error in readInterface");
            return false;
          }
          if ( !ok ) 
             return false;
             
             
          ParameterizedClassInterface pci = parser.getParameterizedClassInterface();
          if ( pci != null ) {
                 // found a parameterized class
              pci.setClassFile(cf);
              if ( pcl.put( pci.getName(), pci ) != null ) {
                 error.show("Duplicate class: " + pci.getName());
              }
              return false;
          }
          else {
             ClassTypeInterface cti = parser.getInterface();
             if ( cti == null || cti.getName() == null ) 
                error.show("Internal error at readInterface");
             cti.setClassFile(cf);
             cf.setClassTypeInterface(cti);
             if ( til.add( cti ) != null ) {
                error.show("Duplicate class : " + cti.getName());
                return false;
             }
          }
       }
       return true;
    }

      
      
      
      
    private void createClassesFromClassList() {
        ToCreateClassList list = toCreateClassList;
        ToCreateClassList all = list;
        toCreateClassList = new ToCreateClassList();
        Hashtable allArrays = new Hashtable();
        
        while ( list.getSize() != 0 || toCreateArrayList.getSize() != 0 ) {
            Hashtable t = list.getTable();
            Enumeration e = t.keys();
            ClassType classType;
            while ( e.hasMoreElements() ) {
                String javaName = (String ) e.nextElement();
                classType = (ClassType ) t.get( javaName );
                createInstantiatedClass(classType);
            }
            e = toCreateArrayList.keys();
            while ( e.hasMoreElements() ) {
                String arrayName = (String ) e.nextElement();
                
                if ( allArrays.get(arrayName) == null ) {
                   ArrayType anArray = toCreateArrayList.get(arrayName);
                   createArray(anArray);
                   allArrays.put(arrayName, "");
                }
            }
               // check if the new classes to be created have not been created before
            Vector removeList = new Vector();
            t = toCreateClassList.getTable();
            e = t.keys();
            while ( e.hasMoreElements() ) {
                String javaName = (String ) e.nextElement();
                classType = (ClassType ) t.get( javaName );
                if ( all.get(javaName) == null )
                   // class have not been created before
                   all.add(classType);
                else 
                   // class have been created before. Don't create it again
                   removeList.addElement(javaName);
            }
            e = removeList.elements();
            while ( e.hasMoreElements() ) 
               toCreateClassList.remove( (String ) e.nextElement());

            
            list = toCreateClassList;
            toCreateArrayList = new ToCreateArrayList();
            toCreateClassList = new ToCreateClassList();
        }
    }


    private boolean createInstantiatedClass( ClassType classType ) 
       /* create a real class from classType, which is something like List(char) or Stack(Person)
          Create .g and .gi files, put in projectDir/classDir */
    {
        ParserParamClass ppc = new ParserParamClass(error, projectDir, classDir, separator);
        ClassFile outcf;
        
        ParameterizedClassInterface pci = (ParameterizedClassInterface ) 
               pcl.get( classType.getName() );
        if ( pci == null ) {
            error.show("Parameterized class " + classType.getName() + " not found");
            return false;
        }
        if ( (outcf = ppc.create(classType, pci)) != null ) 
           if ( readInterface(outcf, false) ) {
               ccl.addElement(outcf);
               return true;
           }
        return false;
    }
          
    
    
    private boolean createAnyArray( ArrayType anArray ) {
        
        if ( anArray.getNumDimensions() != 1 ) {
            error.show("Attempt to use an array of non-basic class with " + 
               anArray.getNumDimensions() + " dimensions" );
            return false;
        }
        String dir;
        String filename;
        
        MyFile mf;
        dir = projectDir + separator + arrayDir + separator;
            
        if ( contentsOf_array_any_1_g == null ) {
            filename = templateDir + separator + "array-any-1.g";
            mf = new MyFile(filename);
            contentsOf_array_any_1_g = mf.readFile();
            if ( mf.getError() != MyFile.ok_e ) {
                error.show("Error reading " + mf.getName() );
                return false;
            }
            fileAny1g = new File(filename);
        }
        
        File f;
        f = new File(dir + anArray.getJavaName());
        if ( f.lastModified() < fileAny1g.lastModified() )
           CreateArray.createArrays( contentsOf_array_any_1_g, dir,
                anArray.getJavaName(), anArray.getType().getName(), null, 
                anArray.getType().getJavaName() );

        ClassFile outcf = new ClassFile(projectDir + separator + arrayDir, anArray.getJavaName() ); 
        if ( readInterface(outcf, false) ) {
            ccl.addElement(outcf);
        }

        
        if ( contentsOf_iter_array_any_1_g == null ) {
            filename = templateDir + separator + "iter-any-1.g";
            mf = new MyFile(filename);
            contentsOf_iter_array_any_1_g = mf.readFile();
            if ( mf.getError() != MyFile.ok_e ) {
                error.show("Error reading " + mf.getName() );
                return false;
            }
            fileIterAny1g = new File(filename);
        }
          
        f = new File( dir + "AnyIter$" + anArray.getType().getJavaName() );
        if ( f.lastModified() < fileIterAny1g.lastModified() )
        
        CreateArray.createArrays( contentsOf_iter_array_any_1_g, dir,
             "AnyIter$" + anArray.getType().getJavaName(), 
             anArray.getType().getName(), null, 
             anArray.getType().getJavaName() );
             
        outcf = new ClassFile(projectDir + separator + arrayDir, 
          "AnyIter$" + anArray.getType().getJavaName() ); 
        if ( readInterface(outcf, false) ) {
            ccl.addElement(outcf);
        }

             
        return true;         
    }
    
    private void createArray( ArrayType anArray ) {

        if ( ! anArray.getType().hasArray() )
            createAnyArray(anArray);
    }
    
    
    private void createAuxArrayClasses() {
        
        for ( int i = 0; i < Type.arrayAnyValue.length; i++ ) {
            TypeList typeList = new TypeList();
            typeList.add( new ClassType( Type.arrayAnyValue[i].getName(), null) );
            toCreateClassList.add( new ClassType( "Iter", typeList ) );
            toCreateClassList.add( new ClassType( "Function", typeList ) );
            toCreateClassList.add( new ClassType( "Command", typeList ) );
            toCreateClassList.add( new ClassType( "Filter", typeList ) );
        }
    }
            

    private void reifyBasicClasses() {
       ClassTypeInterface cti;
       for ( int i = 0; i < Type.arrayAnyValue.length; i++ ) {
           if ( (cti = til.get(Type.arrayAnyValue[i].getName())) == null )
              errorMsg.show("Class " + Type.arrayAnyValue[i].getName() + " not found");
           Type.arrayAnyValue[i].setTypeInterface(cti);
       }
       Type.anyTypeInterface = til.get("Any");
       Type.anyClassTypeInterface = til.get("AnyClass");
       Type.anyClassObjectTypeInterface = til.get("AnyClassObject");
       Type.anyValueTypeInterface = til.get("AnyValue");
       Type.reflectiveAnyClassTypeInterface = til.get("ReflectiveAnyClass");
       Type.nilTypeInterface = til.get("Nil");
       Type.CharTypeInterface = til.get("Char");
       Type.BooleanTypeInterface = til.get("Boolean");
       Type.ByteTypeInterface = til.get("Byte");
       Type.IntegerTypeInterface  = til.get("Integer");
       Type.LongTypeInterface = til.get("Long");
       Type.RealTypeInterface = til.get("Real");
       Type.DoubleTypeInterface = til.get("Double");        
       Type.StringTypeInterface = til.get("String");
       Type.wrapperCharTypeInterface = til.get("Char");
       Type.wrapperByteTypeInterface = til.get("Byte");
       Type.wrapperBooleanTypeInterface = til.get("Boolean");
       Type.wrapperIntegerTypeInterface = til.get("Integer");
       Type.wrapperLongTypeInterface = til.get("Long");
       Type.wrapperRealTypeInterface = til.get("Real");
       Type.wrapperDoubleTypeInterface = til.get("Double");
       
       
       if ( Type.anyTypeInterface == null || 
            Type.anyClassTypeInterface == null ||
            Type.anyClassObjectTypeInterface == null || 
            Type.anyValueTypeInterface == null ||
            Type.reflectiveAnyClassTypeInterface == null ||
            Type.nilTypeInterface == null ||
            Type.CharTypeInterface == null ||
            Type.BooleanTypeInterface == null ||
            Type.ByteTypeInterface == null || 
            Type.IntegerTypeInterface == null ||
            Type.LongTypeInterface == null ||
            Type.RealTypeInterface == null ||
            Type.DoubleTypeInterface == null ||
            Type.StringTypeInterface == null ||
            Type.wrapperCharTypeInterface == null ||
            Type.wrapperByteTypeInterface == null ||
            Type.wrapperBooleanTypeInterface == null ||
            Type.wrapperIntegerTypeInterface == null ||
            Type.wrapperLongTypeInterface == null ||
            Type.wrapperRealTypeInterface == null ||
            Type.wrapperDoubleTypeInterface == null 
            )
            
          errorMsg.show("One of the following classes was not found: Any, AnyClass, ... Char, Boolean, ...");
          
        Type.nilType.setTypeInterface(Type.nilTypeInterface);
        
        Type.anyType.setTypeInterface(Type.anyTypeInterface);
        Type.anyClassType.setTypeInterface(Type.anyClassTypeInterface);
        Type.anyValueType.setTypeInterface(Type.anyValueTypeInterface);
        Type.anyClassObjectType.setTypeInterface(Type.anyClassObjectTypeInterface);
        Type.reflectiveAnyClassType.setTypeInterface(Type.reflectiveAnyClassTypeInterface);
        Type.wrapperCharType.setTypeInterface(Type.wrapperCharTypeInterface);
        Type.wrapperByteType.setTypeInterface(Type.wrapperByteTypeInterface);
        Type.wrapperBooleanType.setTypeInterface(Type.wrapperBooleanTypeInterface);
        Type.wrapperIntegerType.setTypeInterface(Type.wrapperIntegerTypeInterface);
        Type.wrapperLongType.setTypeInterface(Type.wrapperLongTypeInterface);
        Type.wrapperRealType.setTypeInterface(Type.wrapperRealTypeInterface);
        Type.wrapperDoubleType.setTypeInterface(Type.wrapperDoubleTypeInterface);
        
    }
    
     
    private boolean addBasicLib() {
        
        if ( pathBasicLib == null ) {
            error.show("No path for basic library classes was specified", -1);
            return false;
        }
        else 
          return addClasses(pathBasicLib);
    }

    private boolean addWrapperLib() {
        
        if ( pathWrapper != null ) 
           return addClasses(pathWrapper);
        else
           return true;
    }

    private boolean addArrayLib() {

        boolean result = true;
        Enumeration e = arrayPathLib.elements();
        while ( e.hasMoreElements() ) 
           if ( ! addClasses( (String ) e.nextElement() ) ) 
               result = false;
        return result;
    }
    
    private boolean addClasses( String path ) {
        
        File file = new File(path);
        if ( ! file.exists() || ! file.canRead() || ! file.isDirectory() ) {
           error.show("The path " + path + 
              " does not exist or I can't read it or it is not a directory", -1);
           return false;
        }
        else {
           String allFiles[] = file.list();
           Hashtable fileNames = new Hashtable();
           for( int i = 0; i < allFiles.length; i++ ) 
              if ( allFiles[i].endsWith(".g") || allFiles[i].endsWith(".gi") ) 
                  fileNames.put( 
                     allFiles[i].substring(0, allFiles[i].lastIndexOf('.')), "" );
           Enumeration e = fileNames.keys();
           while ( e.hasMoreElements() ) {
              String name = (String ) e.nextElement();
              cfl.addElement( new ClassFile( path, name ) );
           }
          return true;
       }
    }
    
    



    private void parse() {
        Vector list = fileNameList();
        optionsList();
        if ( list.size() < 1 ) 
           error.show("There must be at least one class in the program, besides the libraries");
        else {
           startClass = (String ) list.elementAt(0);
           if ( startClass.indexOf('.') >= 0 || 
                startClass.indexOf(separator) >= 0 )
              error.show("The first line in the project should contain a class name, not a file name");
        }
        Enumeration e = list.elements();
        while ( e.hasMoreElements() ) 
           cfl.addElement( new ClassFile(projectDir, (String ) e.nextElement()) );
        if ( token.tk != PSymbol.EOF )
          error.show("EOF expected", lineNumber);
        if ( templateDir == null ) 
           error.show("templateDir not specified in project");
    }
    
    private Vector fileNameList() {
        Vector list = new Vector();
        while ( token.tk == PSymbol.STRING || token.tk == PSymbol.PATHNAME ) {
            String name  = (String ) token.getValue();
            int i = name.lastIndexOf('.');
            if ( i >= 0 ) 
                name = name.substring(0, i);
            list.addElement(name);
            nextToken();
        }
        return list;
    }
            
    private void optionsList() {
        if ( token.tk != PSymbol.OPTIONS ) 
          error.show("Options expected", lineNumber);
        else {
            inFileName = false;
            nextToken();
            while ( token.tk != PSymbol.EOF )
              option();
        }
    }

    
    private void option() {
        switch ( token.tk ) {
            case PSymbol.BASICLIB :
              nextToken();
              if ( token.tk != PSymbol.ASSIGN )
                error.show("= expected", lineNumber);
              else {
                inFileName = true;
                nextToken();
                inFileName = false;
                if ( token.tk != PSymbol.STRING && token.tk != PSymbol.PATHNAME )
                  error.show("File name expected", lineNumber);
                else {
                    pathBasicLib = (String ) token.getValue();
                    nextToken();
                }
              }
              break;
            case PSymbol.LIB :
              nextToken();
              if ( token.tk != PSymbol.ASSIGN )
                error.show("= expected", lineNumber);
              else {
                inFileName = true;
                nextToken();
                inFileName = false;
                if ( token.tk != PSymbol.STRING && token.tk != PSymbol.PATHNAME )
                  error.show("File name expected", lineNumber);
                else {
                    arrayPathLib.addElement( (String ) token.getValue() );
                    nextToken();
                }
              }
              break;
            case PSymbol.JAVALIB :
              nextToken();
              if ( token.tk != PSymbol.ASSIGN )
                error.show("= expected", lineNumber);
              else {
                inFileName = true;
                nextToken();
                inFileName = false;
                if ( token.tk != PSymbol.STRING && token.tk != PSymbol.PATHNAME )
                  error.show("File name expected", lineNumber);
                else {
                    arrayJavaLib.addElement( (String ) token.getValue() );
                    nextToken();
                }
              }
              break;
            case PSymbol.WRAPPER:
              nextToken();
              if ( token.tk != PSymbol.ASSIGN )
                error.show("= expected", lineNumber);
              else {
                inFileName = true;
                nextToken();
                inFileName = false;
                if ( token.tk != PSymbol.STRING && token.tk != PSymbol.PATHNAME )
                  error.show("File name expected", lineNumber);
                else {
                    pathWrapper = (String ) token.getValue();
                    nextToken();
                }
              }
              break;
            case PSymbol.TEMPLATEDIR : 
              nextToken();
              if ( token.tk != PSymbol.ASSIGN )
                error.show("= expected", lineNumber);
              else {
                inFileName = true;
                nextToken();
                inFileName = false;
                if ( token.tk != PSymbol.STRING && token.tk != PSymbol.PATHNAME )
                  error.show("File name expected", lineNumber);
                else {
                    templateDir = (String ) token.getValue();
                    /* if ( templateDir.length > 0 && 
                         templateDir.charAt(templateDir.length - 1) != separator.charAt(0) )
                       templateDir = templateDir + separator; */
                    nextToken();
                }
              }
              break;
            case PSymbol.PROJECTDIR: 
              nextToken();
              if ( token.tk != PSymbol.ASSIGN )
                error.show("= expected", lineNumber);
              else {
                inFileName = true;
                nextToken();
                inFileName = false;
                if ( token.tk != PSymbol.STRING && token.tk != PSymbol.PATHNAME )
                  error.show("File name expected", lineNumber);
                else {
                    projectDir = (String ) token.getValue();
                    nextToken();
                    Enumeration e = toBeCompiledList.elements();
                    while ( e.hasMoreElements() ) {
                       ClassFile cf = (ClassFile ) e.nextElement();
                       cf.setDirectory(projectDir);
                    }
                }
              }
              break;
            case PSymbol.USEIRL :
              useIRL = true;
              nextToken();
              break;
            case PSymbol.COMPILEONLY :
              compileOnly = true;
              nextToken();
              break;
            case PSymbol.COMPILEBASICCLASSES :
              compileBasicClasses = true;
              nextToken();
              break;
            case PSymbol.ARRAYDIR :
              nextToken();
              if ( token.tk != PSymbol.ASSIGN )
                error.show("= expected", lineNumber);
              else {
                inFileName = true;
                nextToken();
                inFileName = false;
                if ( token.tk != PSymbol.STRING && token.tk != PSymbol.PATHNAME )
                  error.show("Directory name expected", lineNumber);
                else {
                    arrayDir = (String ) token.getValue();
                    nextToken();
                }
              }
              break;
            case PSymbol.CLASSDIR:
              nextToken();
              if ( token.tk != PSymbol.ASSIGN )
                error.show("= expected", lineNumber);
              else {
                inFileName = true;
                nextToken();
                inFileName = false;
                if ( token.tk != PSymbol.STRING && token.tk != PSymbol.PATHNAME )
                  error.show("Directory name expected", lineNumber);
                else {
                    classDir = (String ) token.getValue();
                    nextToken();
                }
              }
              break;
            case PSymbol.INTERFACEDIR :
              nextToken();
              if ( token.tk != PSymbol.ASSIGN )
                error.show("= expected", lineNumber);
              else {
                inFileName = true;
                nextToken();
                inFileName = false;
                if ( token.tk != PSymbol.STRING && token.tk != PSymbol.PATHNAME )
                  error.show("Directory name expected", lineNumber);
                else {
                    interfaceDir = (String ) token.getValue();
                    nextToken();
                }
              }
              break;
            case PSymbol.SUBTYPEEQUALSUBCLASS : 
               subtypeEqualSubclass = true;
               break;
            default:
              error.show("Option expected", lineNumber);
              
        }
    }
              
        
        
        
    
    private void nextToken() {
        char ch;
        
        while (  (ch = input[tokenPos]) == ' ' || ch == '\r' ||
                 ch == '\t' || ch == '\n')  {
            // count the number of lines
          if ( ch == '\n')
            lineNumber++;
          tokenPos++;
          }
        if ( ch == '\0') 
          token = new PSymbol(PSymbol.EOF);
        else
          if ( input[tokenPos] == '/' && input[tokenPos + 1] == '/' ) {
                // comment found
               while ( input[tokenPos] != '\0'&& input[tokenPos] != '\n' )
                 tokenPos++;
               nextToken();
               }
          else {
            if ( inFileName ) {
                if ( ch == '\"' ) {
                    tokenPos++;
                    getString();
                }
                else {
                    StringBuffer ident = new StringBuffer();
                    while ( ! Character.isWhitespace(input[tokenPos]) 
                            && input[tokenPos] != '\0' ) {
                        ident.append( input[tokenPos] );
                        tokenPos++;
                    }
                    String s = ident.toString();
                    if ( beforeOptions && s.toLowerCase().compareTo("options") == 0 ) {
                        beforeOptions = false;
                        token = new PSymbol( PSymbol.OPTIONS );
                    }
                    else
                        token = new PSymbol( PSymbol.PATHNAME, s );
                    }
            }
            else {

               if ( Character.isLetter( ch ) ) {
                   // get a keyword
                   StringBuffer ident = new StringBuffer();
                   while ( Character.isLetter( input[tokenPos] ) ) {
                       ident.append(input[tokenPos]);
                       tokenPos++;
                   }
                   String identStr = ident.toString().toLowerCase();
                      // if identStr is in the list of keywords, it is a keyword !
                   Object value = keywordsTable.get(identStr);
                   if ( value == null ) 
                     error.show( "Option " + ident + " is not valid", lineNumber );
                   else 
                       // get the token
                     token = (PSymbol ) value;
                   if ( Character.isDigit(input[tokenPos]) )
                     error.show("Word followed by a number", lineNumber);
               }
               else if ( Character.isDigit( ch ) ) {
                   // get a number
                   StringBuffer number = new StringBuffer();
                   while ( Character.isDigit( input[tokenPos] ) ) {
                       number.append(input[tokenPos]);
                       tokenPos++;
                   }
                   token =  new PSymbol( PSymbol.NUMBER, new Integer(number.toString()) );
               } else {
                   tokenPos++;
                   switch ( ch ) {
                       case '=' :
                         token = new PSymbol( PSymbol.ASSIGN );
                         break;
                       case '"' :
                          getString();
                          break;
                       default :
                         error.show("Invalid Character: '" + ch + "'", lineNumber);
                       }
                   }
               }
             }
    }

     
    private void getString() {
        
        StringBuffer ident = new StringBuffer();
        while ( input[tokenPos] != '\"' && input[tokenPos] != '\0' ) {
           ident.append(input[tokenPos]);
           tokenPos++;
        }
        String identStr = ident.toString();
        token = new PSymbol( PSymbol.STRING, identStr );
        if ( input[tokenPos] == '\0' ) 
           error.signal("End-of-file unexpected", lineNumber);
        else
           tokenPos++;
    }
    

      // contains the keywords
    static private Hashtable keywordsTable;

     // this code will be executed only once for each program execution
     static {
        keywordsTable = new Hashtable();
        keywordsTable.put( "options", new PSymbol(PSymbol.OPTIONS) );
        keywordsTable.put( "basiclib", new PSymbol(PSymbol.BASICLIB) );
        keywordsTable.put( "lib", new PSymbol(PSymbol.LIB) );
        keywordsTable.put( "wrapper", new PSymbol(PSymbol.WRAPPER) );
        keywordsTable.put( "useirl", new PSymbol(PSymbol.USEIRL) );
        keywordsTable.put( "compileonly", new PSymbol(PSymbol.COMPILEONLY) );
        keywordsTable.put( "compilebasicclasses", new PSymbol(PSymbol.COMPILEBASICCLASSES) );
        keywordsTable.put( "arraydir", new PSymbol(PSymbol.ARRAYDIR) );
        keywordsTable.put( "classdir", new PSymbol(PSymbol.CLASSDIR) );
        keywordsTable.put( "interfacedir", new PSymbol(PSymbol.INTERFACEDIR) );
        keywordsTable.put( "templatedir", new PSymbol(PSymbol.TEMPLATEDIR) );
        keywordsTable.put( "projectdir", new PSymbol(PSymbol.PROJECTDIR) );
        keywordsTable.put( "subtypeequalsubclass", new PSymbol(PSymbol.SUBTYPEEQUALSUBCLASS) );
     }
     
    static private Hashtable basicClassesTable;
    static {
        basicClassesTable = new Hashtable();
        basicClassesTable.put("char", "");
        basicClassesTable.put("boolean", "");
        basicClassesTable.put("byte", "");
        basicClassesTable.put("integer", "");
        basicClassesTable.put("long", "");
        basicClassesTable.put("real", "");
        basicClassesTable.put("double", "");
        basicClassesTable.put("String", "");
    }
        
     
     
      // current token
    private PSymbol token;
    private int  tokenPos, lineNumber;
      // program given as input - source code
    private char []input;
    private String name, projectDir, separator;
    private ProjectError error;
    private ErrorMsg errorMsg;
    private SymbolTable symbolTable;

    private String pathBasicLib; // basic classes and IRL
    private String pathWrapper;  // Wrapper classes
    
    private boolean useIRL, compileOnly, compileBasicClasses;
    
    private int state;
    
    private Vector arrayPathLib, arrayJavaLib, toBeCompiledList;
    //c private ClassList classList;
    private String arrayDir, classDir, templateDir, interfaceDir;
    char []contentsOf_array_any_1_g;
    char []contentsOf_iter_array_any_1_g;
    private ToCreateClassList toCreateClassList;
    private ToCreateArrayList toCreateArrayList;
    private boolean beforeOptions, inFileName;

    private Vector dirList;  // list of all directories that will contain java files 
    private Vector cfl; // classFileList
    private Vector ccl; // compileClassList
    private Hashtable pcl; // parameterizedClassList. Classes that have not been compiled
    private TypeInterfaceList til; // typeInterfaceList
    
    private File fileAny1g; // one of the templates for arrays
    private File fileIterAny1g; // other !
    
    public static boolean subtypeEqualSubclass;
    
    private String startClass; // class in which the execution should begin
}
    
