package Interface;

import java.util.*;
import java.lang.*;
import CompilerError.*;
import MyLexer;
import MySym;
import AST.*;
import java.io.*;
import SymbolTable;


/*

InterfaceDec ::=  [ ClassObjectDec ] ClassDec ;

ClassObjectDec ::= "object" IDENT 
        [ MethodInterface ] 
        [ ClassObjectPublicPart ] 
        [ ClassObjectPrivatePart ]
        "end" ;

ClassobjectPublicPart ::=  "public" ":"  { MethodInterface | ConstDecList }*  ;

ClassObjectPrivatePart ::= "private" ":" 
    { MethodInterface | ConstDecList | "var" ClassObjectVariableDecList }* ;


MethodInterface ::= [ "abstract" ] "proc" IDENT "(" [ FormalParamDecList ] ")" [ ":" Type ]
    AssertClause ;
    
ConstDecList ::= "const" ConstIdList ";"  ;
ConstIdList ::=  IDENT ":" Type { "," IDENT ":" Type }* ;

FormalParamDecList ::= FormalParamDec { ";" FormalParamDec }
FormalParamDec ::= IDENT { "," IDENT } ":" Type

ClassObjectVariableDecList ::= ClassObjectVariableDec { ";" ClassObjectVariableDec }
ClassObjectVariableDec ::= IDENT { "," IDENT } ":" Type

ClassDec ::= [ "abstract" ] "class" IDENT 
      ClassConstructorList 
      [ ClassPublicPart ] 
      [ ClassPrivatePart ]
      "end" ;

ClassConstructorList ::= { MethodInterface } ;
ClassPublicPart ::=  "public" ":" { MethodInterface }+ ;

ClassPrivatePart ::= "private" ":" 
    { MethodInterface | "var" InstanceVariableDecList }* ;

InstanceVariableDecList ::= InstanceVariableDec { ";" InstanceVariableDec }
InstanceVariableDec ::= IDENT { "," IDENT } ":" Type

*/

public class InterfaceParser {
    

   public InterfaceParser( 
        MyLexer lexer, 
        ErrorMsg errorMsg,
        ToCreateClassList toCreateClassList,
        ToCreateArrayList toCreateArrayList,
        boolean is_gi
        ) {
            
      this.errorMsg = errorMsg;
      this.lexer = lexer;
      this.toCreateClassList = toCreateClassList;
      this.toCreateArrayList = toCreateArrayList;
      this.is_gi = is_gi;
      parameterizedClassInterface = null;
   }
   
    
    
    private ClassTypeInterface        currentInterface;
    private MethodInterface           currentMethodInterface;
    private ClassObjectTypeInterface  currentClassObjectInterface;   

    private ParameterizedClassInterface parameterizedClassInterface;
    private ToCreateClassList toCreateClassList;
    private ToCreateArrayList toCreateArrayList;
    private boolean is_gi;
  
    private MyLexer               lexer;
    private ErrorMsg              errorMsg;
    private Hashtable             methodTable, variableConstTable, localTable; 

    private int currentMethodNumber;
    private int lastMethodNumber;
    
    private Hashtable classObjectMethodTable, classObjectVariableConstTable;


   
   public ClassTypeInterface getInterface() {
      return currentInterface;
   }


   public ParameterizedClassInterface getParameterizedClassInterface() {
      return parameterizedClassInterface;
   }
   
   
   private Type basicType() {
      switch ( lexer.token ) {
         case MySym.BOOLEAN :
            return Type.booleanType;
         case MySym.BYTE :
            return Type.byteType;
         case MySym.CHAR :
            return Type.charType;
         case MySym.INTEGER :
            return Type.integerType;
         case MySym.LONG:
            return Type.longType;
         case MySym.REAL :
            return Type.realType;
         case MySym.DOUBLE :
            return Type.doubleType;
         default :
            return null;
      }
   }


   private ClassType parse_class() {
    
      TypeList aTypeList = null;
      String ident;
      
      if ( lexer.token != MySym.IDENT ) {
         errorMsg.show("Identifier expected");
         ident = lexer.ident;
      }
      else {
         ident = lexer.ident;
         lexer.next();
      }
      if ( lexer.token == MySym.LEFTPAR ) {
         lexer.next();
         if ( lexer.token == MySym.EXCEPTION ) 
            errorMsg.show("Probably exception clause put after a method return type");
         aTypeList = typeList();
         if ( lexer.token != MySym.RIGHTPAR ) 
             errorMsg.show(") expected");
         else
            lexer.next();
      }
          
      ClassType aClass;
      aClass = new ClassType(ident, aTypeList);
      aClass.setRealJavaName();
      if ( aTypeList != null ) {
         toCreateClassList.add(aClass);
      }
      return aClass;
   }




   public boolean parse() {
    
      
      try {
         classDec();
      } catch ( CompilerException e ) {
         return false;
      }
      return true;
   }
   
   
      

   private void classDec() {
       // InterfaceDec ::=  [ ClassObjectDec ] ClassDec ;
    
      boolean withClassObject = false;
      MethodInterface m;
      MethodInterfaceList methodList;
      boolean isBasicType = false;
      boolean isShellClass = false;


      currentClassObjectInterface = new ClassObjectTypeInterface();
      currentInterface = new ClassTypeInterface();
      currentInterface.setClassObjectInterface(currentClassObjectInterface);
      currentClassObjectInterface.setClassTypeInterface(currentInterface);
         
      lastMethodNumber = 0;
      
      if ( lexer.token == MySym.SHELL ) {
         lexer.next();
         currentInterface.setIsShellClass(true);
         isShellClass = true;
      }
      else {
         if ( lexer.token == MySym.OBJECT ) {
            if ( classObjectDec() ) 
               return;
            withClassObject = true;
         }
      
         if ( lexer.token == MySym.ABSTRACT ) {
            currentInterface.setIsAbstract(true);
            lexer.next();
         }
         if ( lexer.token == MySym.REFLECTIVE ) {
            currentInterface.setIsReflective(true); 
            lexer.next();
         }
      }
      
      methodTable = new Hashtable();
      variableConstTable = new Hashtable();

      if ( lexer.token != MySym.CLASS ) {
         if ( ! withClassObject ) 
            errorMsg.show("'class' expected");
      }
      else {
         lexer.next();
         
         if ( lexer.token != MySym.IDENT ) {
            if ( basicType() == null ) 
               errorMsg.show("Class name expected");
            else
               isBasicType = true;
         }
         if ( currentInterface.getName() != null && 
              currentInterface.getName().compareTo(lexer.ident) != 0 )
            errorMsg.show("Class object and class name are not equal");
         currentClassObjectInterface.setName(lexer.ident);
         currentInterface.setName(lexer.ident);
         
         currentClassObjectInterface.setIsBasicType(isBasicType);
         currentInterface.setIsBasicType(isBasicType);
         lexer.next();
   
              // normal class
         if ( lexer.token != MySym.LEFTPAR ) {
            if ( isShellClass ) 
               errorMsg.show("( expected");
         }
         else 
            if ( isShellClass ) {
               lexer.next();
               ClassType ct = parse_class();
               currentInterface.setAttachedType(ct);
               if ( lexer.token != MySym.RIGHTPAR )
                  errorMsg.show(") expected");
               lexer.next();
               /*
               String shellName = NameServer.getShellClassName(
                   currentInterface.getName(), ct);
               currentClassObjectInterface.setName(shellName);
               currentInterface.setName(shellName);
               */
            }
            else {
               parameterizedClassInterface = new ParameterizedClassInterface(
                    currentInterface.getName() );
               classParamList();
               return;
            }
            
         if ( lexer.token == MySym.SUBCLASSOF ) {
            lexer.next();
            ClassType superclass = parse_class();
            currentInterface.setSuperclassName(superclass);
            if ( currentInterface.getIsReflective() &&
                 superclass.getName().compareTo("ReflectiveAnyClass") != 0 ) 
               errorMsg.show("To this date, reflective classes cannot inherit from any class");
         }
         else if ( isShellClass ) 
            currentInterface.setSuperclassName(currentInterface.getAttachedType());
         else if ( currentInterface.getIsReflective() ) 
            currentInterface.setSuperclassName(Type.reflectiveAnyClassType);
         

         
         currentMethodNumber = 0;
   
         MethodInterfaceList initMethodList = new MethodInterfaceList();
         
         //currentClassObjectInterface.setInitMethodList(initMethodList);
            
         while ( lexer.token == MySym.PROC ) {
            m = methodDec(true, currentMethodNumber++, false, Method.subclass_v);
            m.setVisibility(Method.subclass_v);
            initMethodList.add(  m );
         }
         currentInterface.setInitMethodList(initMethodList);
         
         
         methodList = new MethodInterfaceList();
         if ( lexer.token == MySym.PUBLIC ) {
            lexer.next();
            if ( lexer.token != MySym.COLON )
               errorMsg.show(": expected");
            else
               lexer.next();
            while ( lexer.token == MySym.PROC ||
                    lexer.token == MySym.ABSTRACT ) {
               m = methodDec(false, currentMethodNumber++, false, Method.public_v);
               m.setVisibility(Method.public_v);
               methodList.add(m);
            }
         }
         
         
            // add the methods deepClone, shallowClone, shallowCopy, ...
         //currentMethodNumber = addCloneCopyMethodsToClass(methodList, currentMethodNumber);
         currentInterface.setPublicMethodList(methodList);
         
         if ( lexer.token == MySym.SUBCLASS) {
            lexer.next();
            if ( lexer.token != MySym.COLON )
               errorMsg.show(": expected");
            else
               lexer.next();
            methodList = new MethodInterfaceList();
            while ( lexer.token == MySym.PROC ||
                    lexer.token == MySym.ABSTRACT ) {
               m = methodDec(false, currentMethodNumber++, false, Method.subclass_v);
               m.setVisibility(Method.subclass_v);
               methodList.add(m);
            }
            currentInterface.setSubclassMethodList(methodList);
         }
         if ( lexer.token == MySym.PRIVATE) {
            lexer.next();
            if ( lexer.token != MySym.COLON )
               errorMsg.show(": expected");
            else
               lexer.next();
            methodList = new MethodInterfaceList();
            InstanceVariableList instVarList = new InstanceVariableList();
            int currentVariableNumber = 0;
            while ( lexer.token == MySym.PROC ||
                    lexer.token == MySym.ABSTRACT ||
                    lexer.token == MySym.VAR ) 
                    
               if ( lexer.token == MySym.VAR ) 
                  currentVariableNumber = instVarDecList(instVarList, currentVariableNumber);
               else {
                  m = methodDec(false, currentMethodNumber++, false, Method.subclass_v);
                  m.setVisibility(Method.subclass_v);
                  methodList.add(m);
               }
                  
            currentInterface.setInstanceVariableList(instVarList);
            currentInterface.setPrivateMethodList(methodList);
         }
         
         javaCodeOp();
         if ( lexer.token != MySym.END )
            errorMsg.show("'end' expected");
         else
            lexer.next();
         if ( lexer.token != MySym.EOF )
            errorMsg.show("End of file expected");
      }
      
      String className = currentInterface.getName();
      boolean isAnyValue = className.compareTo("AnyValue") == 0;
         
      if ( className.compareTo("Any") == 0 || isAnyValue )
         currentInterface.setSuperclassName(null);
      else 
         if ( className.compareTo("AnyClass") == 0 )
            currentInterface.setSuperclassName( Type.anyType );
         else if ( currentInterface.getSuperclassName() == null ) 
            currentInterface.setSuperclassName( Type.anyClassType );
            
            
      if ( className.compareTo("String") == 0 )
         currentInterface.setIsJavaClass(true);
     
      if ( ! is_gi ) {
        
           // if not an interface file, add the methods
         
              // if not an interface file, add the methods
         MethodInterfaceList ml = currentInterface.getPublicMethodList();     
         char []text = textToBeAddedToClass(ml);
         currentInterface.setToAddTextPublic(text);
         MyLexer normalLexer = lexer;
         lexer = new MyLexer(text, errorMsg);
         while ( lexer.token == MySym.PROC ) {
            m = methodDec(false, currentMethodNumber++, false, Method.public_v);
            m.setVisibility(Method.public_v);
            ml.add(m);
         }

         //addNewMethodsToClassObject();
         ml = currentClassObjectInterface.getPublicMethodList();
         text = textToBeAddedToClassObject(ml);
         currentClassObjectInterface.setToAddTextPublic(text);

         methodTable = classObjectMethodTable;
         variableConstTable = classObjectVariableConstTable;
         if ( methodTable == null ) 
            methodTable = new Hashtable();
         if ( variableConstTable == null )
            variableConstTable = new Hashtable();
         currentMethodNumber = lastMethodNumber;
         lexer = new MyLexer(text, errorMsg);
         while ( lexer.token == MySym.PROC ) {
            m = methodDec(false, currentMethodNumber++, true, Method.public_v);
            m.setVisibility(Method.public_v);
            ml.add(m);
         }
         lexer = normalLexer;
         
      }
         
   }
   
   
   
   private void classParamList() {
    
      parameterizedClassInterface.setStartClassParam(lexer.getCurrentChar() - 1);
    
      /* if ( currentInterface.getParamList() != null ) 
         errorMsg.show("Redeclaration of class parameters");
      */
      
        // lexer.token == MySym.LEFTPAR 
      lexer.next();
      Vector paramList = new Vector();
      ClassParameter p;
      
      p = classParam();
      paramList.addElement(p);
      while ( lexer.token == MySym.COMMA ) {
         lexer.next();
         p = classParam();
         paramList.addElement(p);
      }
      parameterizedClassInterface.setParamList(paramList);
      if ( lexer.token != MySym.RIGHTPAR ) 
         errorMsg.show(") expected");
      else {
         parameterizedClassInterface.setEndClassParam(lexer.getCurrentChar());
         lexer.next();
      }
   }
   
   
   private ClassParameter classParam() {
      String ident;
      ClassType paramType;
      
      if ( lexer.token != MySym.IDENT ) {
         errorMsg.show("Identifier expected");
         ident = "Nameless";
      }
      else {
         ident = lexer.ident;
         lexer.next();
      }
      if ( lexer.token == MySym.COLON ) {
         lexer.next();
         paramType = parse_class();
      }
      else
         paramType = null;
      return new ClassParameter(ident, paramType);
   }

   
   
   
   private boolean classObjectDec() {
        // return true if this is a parameterized class
        
         // lexer.token == MySym.OBJECT
      lexer.next();
      
      MethodInterface m;
      boolean inError;
      String ident;
      MethodInterfaceList methodList;
      ConstList constList;
      
      currentMethodNumber = 0;
      
      if ( lexer.token != MySym.IDENT && basicType() == null ) {
         errorMsg.show("Class object name expected");
         return false;
      }
      else {
         ident = lexer.ident;
         lexer.next();
      }
      currentClassObjectInterface.setName(ident);
      currentInterface.setName(ident);

           // normal class
      if ( lexer.token == MySym.LEFTPAR ) {
         parameterizedClassInterface = new ParameterizedClassInterface(
              currentInterface.getName() );
         classParamList();
         return true;
      }
         
      
      classObjectMethodTable = methodTable = new Hashtable();
      classObjectVariableConstTable = variableConstTable = new Hashtable();
      
      if ( lexer.token == MySym.PROC ) {
         m = methodDec(true, currentMethodNumber++, true, Method.private_v);
         currentClassObjectInterface.setInitMethod(m);
         m.setVisibility(Method.private_v);
      }

         
      methodList = new MethodInterfaceList();
      if ( lexer.token == MySym.PUBLIC ) {
         lexer.next();
         if ( lexer.token != MySym.COLON )
            errorMsg.show(": expected");
         else
            lexer.next();
         constList = new ConstList();
         inError = false;
         do {
            switch ( lexer.token ) {
               case MySym.CONST :
                  constDec(constList, Method.public_v);
                  break;
               case MySym.ENUM :
                  enumDec(constList, Method.public_v);
                  break;
               case MySym.ABSTRACT :
                  errorMsg.show("Class object methods cannot be abstract");
                  break;
               case MySym.PROC :
                  m = methodDec(false, currentMethodNumber++, true, Method.public_v);
                  m.setVisibility(Method.public_v);
                  methodList.add(m);
                  break;
               default :
                  inError = true;
            }
         } while ( lexer.token == MySym.PROC ||
                   lexer.token == MySym.ABSTRACT ||
                   lexer.token == MySym.CONST ||
                   lexer.token == MySym.ENUM  );

         currentClassObjectInterface.setPublicConstList(constList);
      }
      //currentMethodNumber = addCloneCopyMethodsToClassObject(methodList, currentMethodNumber);
      currentClassObjectInterface.setPublicMethodList(methodList);
      if ( lexer.token == MySym.PRIVATE ) {
         lexer.next();
         if ( lexer.token != MySym.COLON )
            errorMsg.show(": expected");
         else
            lexer.next();
            
         ClassObjectVariableList   varList;
            
         methodList = new MethodInterfaceList();
         constList = new ConstList();
         varList = new ClassObjectVariableList();
         inError = false;
         int currentVariableNumber = 0;
         do {
            switch ( lexer.token ) {
               case MySym.VAR :
                  currentVariableNumber = objVarDecList(varList, currentVariableNumber);
                  break;
               case MySym.CONST :
                  constDec(constList, Method.private_v);
                  break;
               case MySym.ENUM :
                  enumDec(constList, Method.private_v);
                  break;
               case MySym.ABSTRACT :
                  errorMsg.show("Class object methods cannot be abstract");
                  break;
               case MySym.PROC :
                  m = methodDec(false, currentMethodNumber++, true, Method.public_v);
                  m.setVisibility(Method.public_v);
                  methodList.add(m);
                  break;
               case MySym.JAVACODE : 
               case MySym.JAVAUSE :
                  javaCodeOp();
                  break;
               default : 
                  if ( lexer.token != MySym.END ) {
                     errorMsg.show("Instance variable or method declaration expected");
                     inError = true;
                  }
            }
         } while ( ! inError && lexer.token != MySym.END );
         currentClassObjectInterface.setPrivateMethodList(methodList);
         currentClassObjectInterface.setPrivateConstList(constList);
         currentClassObjectInterface.setClassObjectVariableList(varList);
      }
      if ( lexer.token != MySym.END )
         errorMsg.show("'end' expected");
      else
         lexer.next();
         
      lastMethodNumber = currentMethodNumber;
      return false;
   }


   private void javaCodeOp() {
      
      while ( lexer.token == MySym.JAVAUSE ) {
         lexer.next();
         if ( lexer.token != MySym.LITERALSTRING ) 
            errorMsg.show("String expected");
         else 
            lexer.next();
      }
      if ( lexer.token == MySym.JAVACODE ) {
         lexer.next();
      }
   }
         



   private void constDec( ConstList constList, int visibility ) {
        // lexer.token == MySym.CONST
      lexer.next();
      constList.add( constItem(visibility) );
      while ( lexer.token == MySym.COMMA ) {
         lexer.next();
         constList.add( constItem(visibility) );
      }
      if ( lexer.token != MySym.SEMICOLON )
         errorMsg.show("; expected");
      else
         lexer.next();
   }
   
   
   private Constant constItem(int visibility) {
      String ident;
      Type type;
      
      if ( lexer.token != MySym.IDENT ) {
         errorMsg.show("Identifier expected");
         ident = "nameless";
      }
      else {
         ident = lexer.ident;
         lexer.next();
      }
      if ( variableConstTable.put(ident, "") != null )
         errorMsg.show("Redeclaration of identifier");
       
      type = null;
      if ( lexer.token == MySym.COLON ) {
         lexer.next();
         type = constType();
      }
      if ( lexer.token != MySym.ASSIGN )
         errorMsg.show("= expected");
      else
         lexer.next();
      Expr value = literalConstExpr();
      if ( value == null ) 
         errorMsg.show("Expression expected");
      lexer.next();
      if ( type == null ) 
         return new Constant(ident, value.getType(), value, visibility);
      else
         return new Constant(ident, type, value, visibility);
   }
   
   
   
   
   
   private Type constType() {
      Type result = basicType();
      if ( result == null ) {
         if ( lexer.token != MySym.STRING )
            errorMsg.show("Basic type expected");
         else
            lexer.next();
         result = Type.stringType;
      }
      else
         lexer.next();
      return result;
   }


   private void enumDec( ConstList constList, int visibility ) {
      String ident;
        // lexer.token == MySym.ENUM
      lexer.next();
      if ( lexer.token != MySym.LEFTPAR )
         errorMsg.show("( expected");
      else
         lexer.next();
         
      int currentEnumValue = 0;
      
      while ( true ) {
         if ( lexer.token != MySym.IDENT ) {
            errorMsg.show("Identifier expected");
            ident = "nameless";
         }
         else {
            ident = lexer.ident;
            lexer.next();
         }
         if ( variableConstTable.put(ident, "") != null )
            errorMsg.show("Redeclaration of identifier");
         
         if ( lexer.token == MySym.ASSIGN ) { 
            lexer.next();
            Expr anExpr = literalConstExpr();
            if ( anExpr == null ) 
               errorMsg.show("Expression expected");
            lexer.next();             
            currentEnumValue = ((Integer ) anExpr.getValue()).intValue();
         }
         constList.add( new Constant(ident, Type.integerType, 
            new LiteralIntegerConst(currentEnumValue++), visibility));
         if ( lexer.token == MySym.COMMA ) 
            lexer.next();
         else if ( lexer.token == MySym.RIGHTPAR ) {
            lexer.next();
            if ( lexer.token != MySym.SEMICOLON ) 
                errorMsg.show("; expected");
            else
                lexer.next();
            break;
         }
         else {
            errorMsg.show(", or ) expected");
            break;
         }
      }
   }


   private LiteralBooleanConst booleanConst() {
      switch ( lexer.token ) {
         case MySym.TRUE : 
            return LiteralBooleanConst.booleanTrue; 
         case MySym.FALSE :
            return LiteralBooleanConst.booleanFalse; 
         default :
            return null;
      }
   }

   private LiteralConstExpr literalConstExpr() {
      switch ( lexer.token ) {
         case MySym.TRUE :
         case MySym.FALSE :
            return booleanConst();
         case MySym.BYTECONST :
            return new LiteralByteConst(lexer.byteValue);
         case MySym.CHARCONST :
            return new LiteralCharConst(lexer.charValue);
         case MySym.DOUBLECONST :
            return new LiteralDoubleConst(lexer.doubleValue);
         case MySym.INTEGERCONST :
            return new LiteralIntegerConst(lexer.integerValue);
         case MySym.LONGCONST :
            return new LiteralLongConst(lexer.longValue);
         case MySym.REALCONST :
            return new LiteralRealConst(lexer.realValue);
         default : 
            return null;
      }
   }



   private void exceptionClause() {
         // lexer.token == MySym.LEFTPAR
      lexer.next();
      if ( lexer.token != MySym.EXCEPTION )
         errorMsg.show("'exception' expected");
      else
         lexer.next();
      if ( lexer.token != MySym.COLON )
         errorMsg.show(": expected");
      else
         lexer.next();
      Type t = type();
      if ( lexer.token != MySym.RIGHTPAR )
         errorMsg.show(") expected");
      else
         lexer.next();
      currentMethodInterface.setExceptionType(t);
   }
   

   
   private void formalParamDecList( ParameterList methodParamList ) {
    
      boolean lastIsVariableNumber = formalParamDec(methodParamList);
      while ( lexer.token == MySym.SEMICOLON ) {
         if ( lastIsVariableNumber )
            errorMsg.show("There should be no parameters after \"... array(T}[]\"");
         lexer.next();
         lastIsVariableNumber = formalParamDec(methodParamList);
      }
   }

   
   private boolean formalParamDec( ParameterList methodParamList ) {
        // return true if this is a parameter with "..." (for variable number of parameters)
      Vector localList = new Vector();
      boolean isVariableNumber;
      Parameter p;
      
      while ( true ) {
         if ( lexer.token != MySym.IDENT ) {
            errorMsg.show("Identifier expected");
            return false;
         }
         else {
            p = new Parameter(lexer.ident);
            methodParamList.add(p);
            localList.addElement(p);
            if ( localTable.put(lexer.ident, "") != null ) 
               errorMsg.show("Redeclaration of identifier");
            lexer.next();
         }
         if ( lexer.token == MySym.COMMA ) 
            lexer.next();
         else 
            break;
      }
      if ( lexer.token != MySym.COLON )
         errorMsg.show(": expected");
      else
         lexer.next();
      if ( lexer.token == MySym.THREEPERIOD ) {
         isVariableNumber = true;
         lexer.next();
      }
      else
         isVariableNumber = false;
      Type t = type();
      Enumeration e = localList.elements();
      while ( e.hasMoreElements() )
         ((Parameter ) e.nextElement()).setType(t);
         
      if ( isVariableNumber ) 
        if ( localList.size() != 1 )
          errorMsg.show("... used with more than one parameter");
        else 
          ((Parameter ) localList.elementAt(0)).setIsVariableNumber(true);

      return isVariableNumber;
   }
      
         
      

         

   private int instVarDec( InstanceVariableList instVarList,
                           int currentVariableNumber) {
      Vector localList = new Vector();
      InstanceVariable p;
      while ( true ) {
         if ( lexer.token != MySym.IDENT ) {
            errorMsg.show("Identifier expected");
            return currentVariableNumber;
         }
         else {
            p = new InstanceVariable(lexer.ident, currentVariableNumber++);
            instVarList.add(p);
            localList.addElement(p);
            
            if ( variableConstTable.put(lexer.ident, "") != null )
               errorMsg.show("Redeclaration of identifier");
            lexer.next();
         }
         if ( lexer.token == MySym.COMMA ) 
            lexer.next();
         else 
            break;
      }
      if ( lexer.token != MySym.COLON )
         errorMsg.show(": expected");
      else
         lexer.next();
      Type t = type();
      if ( lexer.token != MySym.SEMICOLON )
         errorMsg.show("; expected");
      else
         lexer.next();
      
      Enumeration e = localList.elements();
      while ( e.hasMoreElements() ) {
         p = (InstanceVariable ) e.nextElement();
         p.setType(t);
      }
      return currentVariableNumber;
   }
      

   private int instVarDecList( InstanceVariableList instVarList,
                                int currentVariableNumber) {
        // lexer.token == MySym.VAR
      lexer.next();
      currentVariableNumber = instVarDec(instVarList, currentVariableNumber);
      while ( lexer.token == MySym.IDENT ) 
         currentVariableNumber = instVarDec(instVarList, currentVariableNumber);
      return currentVariableNumber;
   }





   private int objVarDecList( ClassObjectVariableList varList,
                              int currentVariableNumber) {
        // lexer.token == MySym.VAR
      lexer.next();
      while ( true ) {
         currentVariableNumber = objVarDec(varList, currentVariableNumber);
         if ( lexer.token != MySym.IDENT ) 
            break;
      }
      return currentVariableNumber;
   }
   
   
   
   private int objVarDec( ClassObjectVariableList varList,
                          int currentVariableNumber) {
      ClassObjectVariable v;
      Expr anExpr;
      Vector localList = new Vector();
      
      while ( true ) 
         if ( lexer.token != MySym.IDENT ) {
            errorMsg.show("Identifier expected");
            break;
         }
         else {
            v = new ClassObjectVariable(lexer.ident, currentVariableNumber++);
            localList.addElement(v);
            if ( variableConstTable.put(lexer.ident, "") != null )
               errorMsg.show("Redeclaration of identifier");
            lexer.next();
            if ( lexer.token == MySym.COMMA ) 
               lexer.next();
            else
               break;
         }
      if ( lexer.token != MySym.COLON )
         errorMsg.show(": expected");
      else
         lexer.next();
      Type t = type();
      if ( lexer.token == MySym.ASSIGN ) {
         lexer.next();
         anExpr = literalConstExpr();
         if ( anExpr == null ) {
            errorMsg.show("Expression expected");
         }
         lexer.next();
      }
      else
         anExpr = null;
      if ( lexer.token != MySym.SEMICOLON )
         errorMsg.show("; expected");
      else
         lexer.next();
      if ( anExpr != null ) 
         if ( localList.size() > 1 )
            errorMsg.show("Attempt to initialize more than one variable");
         else 
            ((ClassObjectVariable ) localList.elementAt(0)).setExpr(anExpr);
         
      Enumeration e = localList.elements();
      while ( e.hasMoreElements() ) {
          ClassObjectVariable iv = (ClassObjectVariable ) e.nextElement();
          iv.setType(t);
          varList.add(iv);
      }
      return currentVariableNumber;
      
   }

   private String parse_operator() {
      switch ( lexer.token ) {
         case MySym.PLUSPLUS :
            return "++";
         case MySym.MINUSMINUS :
            return "--";
         case MySym.BINNOT :
            return "~";
         case MySym.PLUS :
            return "+";
         case MySym.MINUS :
            return "-";
         case MySym.MULT :
            return "*";
         case MySym.DIV :
            return "/";
         case MySym.REMAINDER :
            return "%";
         case MySym.BINAND :
            return "&";
         case MySym.BINOR :
            return "|";
         case MySym.BINXOR :
            return "^";
         case MySym.LEFTSHIFT :
            return "<<";
         case MySym.RIGHTSHIFT :
            return ">>";
         case MySym.EQ :
            return "==";
         case MySym.LT :
            return "<";
         case MySym.GT :
            return ">";
         case MySym.LE :
            return "<=";
         case MySym.GE :
            return ">=";
         case MySym.NEQ :
            return "<>";
         case MySym.AND :
            return "and";
         case MySym.OR :
            return "or";
         case MySym.XOR :
            return "xor";
         case MySym.NOT :
            return "not";
         default :
            return null;
      }
   }
         


   private MethodInterface methodDec( 
         boolean isInitMethod, 
         int currentMethodNumber,
         boolean inClassObject,
         int visibility
         ) {
    
      String methodName;
      boolean isAbstract;
      boolean isOperator;
      
      
      localTable = new Hashtable();

      if ( lexer.token == MySym.ABSTRACT ) {
         isAbstract = true;
         lexer.next();
      }
      else
         isAbstract = false;
         
      if ( lexer.token != MySym.PROC )
         errorMsg.show("'proc' expected");
      else
         lexer.next();
         
      if ( lexer.token != MySym.IDENT ) {
           methodName = parse_operator();
           if ( methodName == null ) {
              errorMsg.show("Identifier expected");
              return null;
           }
           isOperator = true;
           lexer.next();
      }
      else {
         methodName = lexer.ident;
         isOperator = false;
         lexer.next();
      }
      
      if ( isInitMethod ) {
         MethodInterface m;
         currentMethodInterface = m = new MethodInterface("init", false, currentMethodNumber);
         if ( methodName.compareTo("init") != 0 ) 
            errorMsg.show("Methods before the public part should be named 'init'");
      }
      else {
         currentMethodInterface = new MethodInterface(methodName, isAbstract, 
                currentMethodNumber);
      }
      
      
      if ( lexer.token != MySym.LEFTPAR )
         errorMsg.show("( expected");
      else
         lexer.next();

      ParameterList methodParamList = new ParameterList();
      if ( lexer.token != MySym.RIGHTPAR && lexer.token != MySym.COLON
           && lexer.token != MySym.VAR )
         formalParamDecList(methodParamList);
      if ( lexer.token != MySym.RIGHTPAR )
         errorMsg.show(") expected");
      else
         lexer.next();
         
      boolean inInterceptAll = false;
      if ( currentInterface.getIsShellClass() && 
            methodName.compareTo("interceptAll") == 0 && ! inClassObject ) {
         inInterceptAll = true;
         if ( visibility != Method.public_v ) 
            errorMsg.show("interceptAll must be a public method");
         if ( methodParamList.getSize() != 2 )
            errorMsg.show("interceptAll should have two parameters");
         Enumeration f = methodParamList.elements();
         Parameter p = (Parameter ) f.nextElement();
         if ( p.getType().getName().compareTo("ObjectMethodInfo") != 0 )
            errorMsg.show("The first parameter of interceptAll should be of type ObjectMethodInfo");
         p = (Parameter ) f.nextElement();
         if (  p.getType().getName().compareTo("array(Any)[]") != 0 )
            errorMsg.show("The second parameter of interceptAll should be of type array(Any)[]");
         currentInterface.setHasInterceptAll(true);
      }

      currentMethodInterface.setParamList(methodParamList);

      String searchName = TypeInterface.buildMethodSearchName(methodName, methodParamList);
      if ( methodTable.put(searchName, "") != null )
         errorMsg.show("Redefinition of method " + methodName);
      variableConstTable.put(methodName, "");
      
      boolean hadExceptionClause = false;
      if ( lexer.token == MySym.LEFTPAR ) {
         exceptionClause();
         hadExceptionClause = true;
      }
         
      if ( lexer.token == MySym.COLON ) {
         lexer.next();
         Type t = type();
         currentMethodInterface.setReturnType(t);
      }

      if ( inInterceptAll ) {
         if ( currentMethodInterface.getReturnType() == null ||
              currentMethodInterface.getReturnType().getName().compareTo("Any") != 0 )
            errorMsg.show("Return type of method interceptAll must be Any");
      }
      
      if ( lexer.token == MySym.ASSERT ) {
         while ( lexer.token != MySym.EOF && lexer.token != MySym.END )
            lexer.next();
         if ( lexer.token == MySym.EOF ) 
            errorMsg.show("Unexpected end-of-file");
         lexer.next();
      }
      if ( ! isAbstract ) {
         if ( lexer.token == MySym.VAR ) {
            LocalVariableList localVariableList = new LocalVariableList();
            localDec(localVariableList);
            currentMethodInterface.setLocalVariableList(localVariableList);
         }
         if ( lexer.token == MySym.BEGIN ) {
            int count = 1;
            lexer.next();
            while ( count > 0 && lexer.token != MySym.EOF ) {
               if ( lexer.token == MySym.BEGIN || lexer.token == MySym.TRY )
                  count++;
               else if ( lexer.token == MySym.END )
                  count--;
               lexer.next();
            }
            if ( count > 0 && lexer.token == MySym.EOF ) 
               errorMsg.show("Unexpected end of file");
         }
      }
      return currentMethodInterface;
   }
      
   
         

   private void localDec(LocalVariableList localVariableList) {
        // lexer.token == MySym.VAR
      lexer.next();
      while ( true ) {
         localVarDec(localVariableList);
         if ( lexer.token != MySym.IDENT ) 
            break;
      }
   }
   

   private void localVarDec(LocalVariableList localVariableList) {
      LocalVariable localVariable;
      Vector localList = new Vector();
      while ( true ) 
         if ( lexer.token != MySym.IDENT ) {
            errorMsg.show("Identifier expected");
            break;
         }
         else {
            localVariable = new LocalVariable(lexer.ident);
            localVariableList.add(localVariable);
            localList.addElement(localVariable);
            if ( localTable.put(lexer.ident, "") != null ) 
               errorMsg.show("Redeclaration of identifier");
            lexer.next();
            if ( lexer.token == MySym.COMMA ) 
               lexer.next();
            else
               break;
         }
      if ( lexer.token != MySym.COLON )
         errorMsg.show(": expected");
      else
         lexer.next();
      Type t = typeExt();
      if ( lexer.token != MySym.SEMICOLON )
         errorMsg.show("; expected");
      else
         lexer.next();
      Enumeration e = localList.elements();
      while ( e.hasMoreElements() )
         ((LocalVariable ) e.nextElement()).setType(t);
   }



   private Type type() {
      Type t = basicType();
      if ( t != null ) {
         lexer.next();
         return t;
      }
      else 
         switch ( lexer.token ) {
            case MySym.TYPE :
               return typeFun();
            case MySym.ARRAY :
               return arrayType();
            default :
               return parse_class();
         }
   }
   
   private ArrayType arrayType() {
        // lexer.token == MySym.ARRAY 
      lexer.next();
      if ( lexer.token != MySym.LEFTPAR )
         errorMsg.show("( expected");
      else
         lexer.next();
      Type t = type();
      if ( lexer.token != MySym.RIGHTPAR )
         errorMsg.show(") expected");
      else
         lexer.next();
      int n = 0;
      do { 
         if ( lexer.token != MySym.LEFTSB )
            errorMsg.show("[ expected");
         else
            lexer.next();
         if ( lexer.token != MySym.RIGHTSB )
            errorMsg.show("] expected");
         else
            lexer.next();
         n++;
      } while ( lexer.token == MySym.LEFTSB );
      ArrayType at = new ArrayType(t, n);
      at.setRealJavaName();
      toCreateArrayList.add(at);
      return at;

   }


   private Type typeFun() {
        // lexer.token == MySym.TYPE 
      lexer.next();
      String ident;
      
      if ( lexer.token != MySym.LEFTPAR )
         errorMsg.show("( expected");
      else
         lexer.next();
      ClassType classType = parse_class();
      if ( lexer.token != MySym.RIGHTPAR )
         errorMsg.show(") expected");
      else
         lexer.next();
      if ( classType.getTypeList() != null ) 
         return new TypeOfClass(classType);
      else {
            // search for a local variable or parameter with name ident
         Enumeration e;
         Variable v;
         ident = classType.getName();
         e = currentMethodInterface.getLocalVariableList().elements();
         while ( e.hasMoreElements() ) {
            v = (Variable ) e.nextElement();
            if ( v.getName().compareTo(ident) == 0 ) 
               return v.getType();
         }
         e = currentMethodInterface.getParamList().elements();
         while ( e.hasMoreElements() ) {
            v = (Variable ) e.nextElement();
            if ( v.getName().compareTo(ident) == 0 ) 
               return v.getType();
         }
         return new TypeOfClass(classType);
      }
   }


   private Type typeExt() {
      if ( lexer.token == MySym.AT ) {
         lexer.next();
         Type t = type();
         return new AtType(t);
      }
      else 
         return type();
   }
   
   
   private TypeList typeList() {
       TypeList typeList = new TypeList(false);
       
       while ( true ) {
           Type t = type();
           typeList.add(t);
           /*
           if ( t instanceof ClassType ) {
               ClassType ct = (ClassType ) t;
               if ( ct.getIsClassParameter() )
                  typeList.setWithClassParameter(true);
           }
           */
           if ( lexer.token == MySym.COMMA ) 
              lexer.next();
           else
              break;
       }
       return typeList;
   }



   private int addCloneCopyMethodsToClass( MethodInterfaceList methodList,
                                     int currentMethodNumber ) {
      
      MethodInterface m;
      ParameterList methodParamList;
      Parameter p;
      
      m = new MethodInterface("deepClone", false, currentMethodNumber++);
      methodParamList = new ParameterList();
      m.setParamList(methodParamList);
      m.setReturnType( new ClassType(currentInterface.getName(), null) );
      m.setVisibility(Method.public_v);
      if ( methodList.get( m.getSearchName() ) == null ) 
         methodList.add(m);
      
      m = new MethodInterface("shallowClone", false, currentMethodNumber++);
      methodParamList = new ParameterList();
      m.setParamList(methodParamList);
      m.setReturnType( new ClassType(currentInterface.getName(), null) );
      m.setVisibility(Method.public_v);
      if ( methodList.get( m.getSearchName() ) == null ) 
         methodList.add(m);
      
      m = new MethodInterface("shallowCopy", false, currentMethodNumber++);
      methodParamList = new ParameterList();
      p = new Parameter("other");
      p.setType( new ClassType("Any", null) );
      methodParamList.add(p);
      m.setParamList(methodParamList);
      m.setReturnType( Type.booleanType );
      m.setVisibility(Method.public_v);
      if ( methodList.get( m.getSearchName() ) == null ) 
         methodList.add(m);
      
      m = new MethodInterface("shallowEqual", false, currentMethodNumber++);
      methodParamList = new ParameterList();
      p = new Parameter("other");
      p.setType( new ClassType("Any", null) );
      methodParamList.add(p);
      m.setParamList(methodParamList);
      m.setReturnType( Type.booleanType );
      m.setVisibility(Method.public_v);
      if ( methodList.get( m.getSearchName() ) == null ) 
         methodList.add(m);
      
      return currentMethodNumber;
   
   }



   private int addCloneCopyMethodsToClassObject( MethodInterfaceList methodList ) {
      
      MethodInterface m;
      ParameterList methodParamList;
      Parameter p;
      
      m = new MethodInterface("deepClone", false, currentMethodNumber++);
      methodParamList = new ParameterList();
      m.setParamList(methodParamList);
      m.setReturnType( new ClassType(currentInterface.getName(), null) );
      m.setVisibility(Method.public_v);
      if ( methodList.get( m.getSearchName() ) == null ) 
         methodList.add(m);
      
      m = new MethodInterface("shallowClone", false, currentMethodNumber++);
      methodParamList = new ParameterList();
      m.setParamList(methodParamList);
      m.setReturnType( new ClassType(currentInterface.getName(), null) );
      m.setVisibility(Method.public_v);
      if ( methodList.get( m.getSearchName() ) == null ) 
         methodList.add(m);
      
      m = new MethodInterface("shallowCopy", false, currentMethodNumber++);
      methodParamList = new ParameterList();
      p = new Parameter("other");
      p.setType( new ClassType("Any", null) );
      methodParamList.add(p);
      m.setParamList(methodParamList);
      m.setReturnType( Type.booleanType );
      m.setVisibility(Method.public_v);
      if ( methodList.get( m.getSearchName() ) == null ) 
         methodList.add(m);
      
      m = new MethodInterface("shallowEqual", false, currentMethodNumber++);
      methodParamList = new ParameterList();
      p = new Parameter("other");
      p.setType( new ClassType("Any", null) );
      methodParamList.add(p);
      m.setParamList(methodParamList);
      m.setReturnType( Type.booleanType );
      m.setVisibility(Method.public_v);
      if ( methodList.get( m.getSearchName() ) == null ) 
         methodList.add(m);
      
      return currentMethodNumber;
   
   }

   private char []textToBeAddedToClass( MethodInterfaceList mil ) {

         
      StringWriter outString = new StringWriter();
      PrintWriter out = new PrintWriter( outString );
      Parameter p;

      p = new Parameter("aClass");
      p.setType( new ClassType("AnyClassObject", null) );
      ParameterList pl = new ParameterList();
      pl.add(p);
      if ( mil.get( TypeInterface.buildMethodSearchName("isObjectOf", pl)) == null ) {
         out.println("      proc isObjectOf( aClass : AnyClassObject ) : boolean");
         out.println("         begin");
         out.println("         java$begin");
         out.println("         return $classObject == _aClass;");
         out.println("         java$end");
         out.println("         end");
         out.println();
      }

      pl = new ParameterList();
      if ( mil.get( TypeInterface.buildMethodSearchName("getInfo", pl )) == null ) {
         out.println("      proc getInfo() : AnyObjectInfo");
         out.println("         begin");
         out.println("         java$begin");
         
         out.println("         return _ObjectInfo.$classObject._new$_Any$$Array$1$_ObjectInstanceVariableInfo$$Array$1$_ObjectMethodInfo$_ClassInfo( this, OarrayIVI, OarrayMI, $i );");
         out.println("         java$end");
         out.println("         end");
         out.println();
      }

      if ( mil.get( TypeInterface.buildMethodSearchName("getClassInfo", pl)) == null ) {
         out.println("      proc getClassInfo() : ClassInfo");
         out.println("         begin");
         out.println("         java$begin");
         out.println("         return $i;");
         out.println("         java$end");
         out.println("         end");
         out.println();
      }

      if ( mil.get( TypeInterface.buildMethodSearchName("getClassObject", pl)) == null ) {
         out.println("      proc getClassObject() : AnyClassObject");
         out.println("         begin");
         out.println("         java$begin");
         out.println("         return $classObject;");
         out.println("         java$end");
         out.println("         end");
         out.println();
      }

      
      out.print("  ");
      char []output = outString.toString().toCharArray();
      output[output.length - 1] = '\0';
      
//      if ( firstTime < 2  ) {
      if ( currentInterface.getName().compareTo("Long") == 0 ) {
         try { 
            PrintWriter pw = new PrintWriter( new FileOutputStream("c:\\jose\\green\\compiler\\c.g") );
            pw.write(output);
            firstTime++;
            pw.close();
         } catch (Exception e ) {
            throw new CompilerException();
         }
      }
      
      return output;
   }


   private char []textToBeAddedToClassObject( MethodInterfaceList mil ) {

      StringWriter outString = new StringWriter();
      PrintWriter out = new PrintWriter( outString );
      ParameterList pl;
      Parameter p;


      if ( ! currentInterface.getIsJavaClass() && ! currentInterface.getIsAbstract() ) {
         MethodInterfaceList initMethodList = currentInterface.getInitMethodList();
         // add new methods to class object
      
         Enumeration f = initMethodList.anyElements();
         while ( f.hasMoreElements() ) {
            MethodInterface initI = (MethodInterface ) f.nextElement();
         
            out.print("      proc new(");
           
            int size = initI.getParamList().getSize();
            Enumeration g = initI.getParamList().elements();
            while ( g.hasMoreElements() ) {
               p = (Parameter ) g.nextElement();
               out.print( " " + p.getGreenName() + " : " + p.getType().getGreenName() );
               if ( --size > 0 )
                  out.print("; ");
            }
            out.println(") : " + currentInterface.getName() );
           
            out.println("      begin");
            out.println("      java$begin");
         
            out.println("      " + currentInterface.getJavaName() + " r = new " + 
                            currentInterface.getJavaName() + "();" );
                            
            out.print( "      r." + initI.getJavaName() + "(" );
            Enumeration h;
            h = initI.getParamList().elements();
            size = initI.getParamList().getSize();
            while ( h.hasMoreElements() ) {
               p = (Parameter ) h.nextElement();
               out.print( p.getJavaName() );
               if ( --size > 0 )
                  out.print(", ");
            }
            out.println(");");
            out.println("      return r;");
          
         
            out.println("       java$end");
            out.println("       end");
         }
      }
      
        // add cast method
      if ( mil.get( castMethodName ) == null && ! currentInterface.getIsBasicType() &&
           ! currentInterface.getIsJavaClass() &&
           currentInterface.getName().compareTo("AnyValue") != 0 ) {
         
         out.println();
         out.println("      proc cast( x : Any )");
         out.println("               ( exception : CatchTypeErrorException ) : " 
              + currentInterface.getName() );
         out.println("         begin");
         out.println("         java$begin");
         out.println("         if ( _x instanceof " + currentInterface.getJavaName() + " )");
//         out.println("            return (" + currentInterface.getJavaName() + " ) _x;");
         if ( currentInterface.getName().indexOf("$Array") >= 0 ||
              currentInterface.getName().indexOf("array(") >= 0 )
            out.println("            return (" + currentInterface.getJavaName() + " ) _x;");
         else
            out.println("            return (_Any ) _x;");
         out.println("         throw new PackedException(_TypeErrorException.$classObject._new());");
         // out.println("         return null;");
         out.println("         java$end");
         out.println("         end");
         out.println();
         
      }
         

      pl = new ParameterList();
      
      if ( mil.get( TypeInterface.buildMethodSearchName("getAssociateClassInfo", pl)) == null ) {
         out.println("      proc getAssociateClassInfo() : ClassInfo");
         out.println("         begin");
         out.println("         java$begin");
         out.println("         return associateClassInfo;");
         out.println("         java$end");
         out.println("         end");
      }
      
      
      if ( mil.get( TypeInterface.buildMethodSearchName("getInitMethod", pl)) == null ) {
         out.println("      proc getInitMethod() : ObjectInitMethodInfo");
         out.println("         begin");
         out.println("         java$begin");
         out.println("         return initMethodInfo;");
         out.println("         java$end");
         out.println("         end");
      }
      p = new Parameter("aClass");
      p.setType( new ClassType("AnyClassObject", null) );
      pl.add(p);
         
      if ( mil.get( TypeInterface.buildMethodSearchName("getAssociateClassInfo", pl)) == null ) {
         out.println("      proc isObjectOf( aClass : AnyClassObject ) : boolean");
         out.println("         begin");
         out.println("         return false;");
         out.println("         end");
      }
         
      pl = new ParameterList();
      if ( mil.get( TypeInterface.buildMethodSearchName("getInfo", pl)) == null ) {
         out.println("      proc getInfo() : AnyObjectInfo");
         out.println("         begin");
         out.println("         java$begin");
         out.println("         return classObjectInfo;");
         out.println("         java$end");
         out.println("         end");
      }
         
      if ( mil.get( TypeInterface.buildMethodSearchName("toString", pl)) == null ) {
         out.println("      proc toString() : String");
         out.println("         begin");
         out.println("         return \"\";");
         out.println("         end");
      }
         
      p = new Parameter("other");
      p.setType( Type.anyType );
      pl.add(p);
      if ( mil.get( TypeInterface.buildMethodSearchName("equals", pl)) == null ) {
         out.println("      proc equals( other : Any ) : boolean");
         out.println("         begin");
         out.println("         return self == other;");
         out.println("         end");
      }
      
      
      out.print("  ");
      char []output = outString.toString().toCharArray();
      output[output.length - 1] = '\0';
      
/*      if ( firstTime < 2 ) {
      if ( currentInterface.getName().compareTo("Long") == 0 ) {

         try {
            PrintWriter pw = new PrintWriter( new FileOutputStream("c:\\jose\\green\\compiler\\co.g") );
            pw.write(output);
            firstTime++;
            pw.close();
         } catch (Exception e ) {
            throw new CompilerException();
         }
      }
*/      
      return output;
      
   }


   private void addNewMethodsToClassObject() {

      ParameterList pl;
      Parameter p;

      MethodInterfaceList mil;
      mil = currentClassObjectInterface.getPublicMethodList();

      MethodInterfaceList initMethodList = currentInterface.getInitMethodList();
      // add new methods to class object
      Enumeration f = initMethodList.elements();
      while ( f.hasMoreElements() ) {
         MethodInterface initI = initMethodList.get( (String ) f.nextElement() );
         
         if ( mil.get( 
                TypeInterface.buildMethodSearchName( "new", initI.getParamList() ) ) != null ) {
            errorMsg.show("There is a new method in the class object with the same parameters as a class init method");
         }
         MethodInterface newI = new MethodInterface("new", false, currentMethodNumber++);
         Type returnType;
         returnType = new ClassType(currentInterface.getName(), null);
         newI.setReturnType( returnType );
         pl = new ParameterList();
         
         Enumeration g = initI.getParamList().elements();
         while ( g.hasMoreElements() ) {
            p = (Parameter ) g.nextElement();
            Parameter np = new Parameter( p.getName() );
            np.setType(p.getType());
            pl.add(np);
         }
         newI.setParamList(pl);
         mil.add(newI);
         
      }
   }


   static int firstTime = 0;
   static public String castMethodName;
   static {
      ParameterList pl = new ParameterList();
      Parameter p = new Parameter("x");
      p.setType(Type.anyType);
      pl.add(p);
      castMethodName = TypeInterface.buildMethodSearchName("cast", pl ); 
   }


}
