import java.util.*;
import java.io.*;
import AST.*;
import CompilerError.*;


/**
  In directory basiclib should be put arrays of all basic types up to dimension
  three ( array(byte)[], array(byte)[][], array(byte)[][][], ... ), the basic 
  classes (char, byte, integer, ...), all classes used by the 
  array classes (Iter$Array$1$byte, ...).
      In directory exceptions should be put all exception classes. 
      In directory greenlib should be  put the basic green library (Vector, Filter, ...)
      
  Grammar for the project
    Project ::= FileNameList  OPTIONS OptionsList 
    FileNameList ::= | FileNameList FileName
    OptionsList ::= Option | OptionsList Option
    Option ::= BASICLIB "=" FileName | LIB "=" FileName | WRAPPER "=" FileName 
              | USEIRL | COMPILEONLY | COMPILEBASICCLASSES 
              | ARRAYDIR "=" FileName | CLASSDIR "=" FileName
              | JAVALIB "=" FileName
              | TEMPLATEDIR  "=" FileName
              | PROJECTDIR = FileName
    FileName ::= STRING | PATHNAME
    
*/


public class GreenProject {
    
    public boolean init( String fileName, ProjectError error,  ErrorMsg errorMsg  ) {
        this.errorMsg = errorMsg;
        this.error = error;
        input = readFile(fileName);
        if ( input == null ) 
           return false;
        setDefaultOptions();
        arrayPathLib = new Vector();
        arrayJavaLib = new Vector();
        classList = new ClassList();
        toBeCompiledList = new Vector();
        toCreateClassList = new ToCreateClassList();
        toCreateArrayList = new ToCreateArrayList();
        separator = File.separator;
        projectDir = "";
        return true;
    }
    
    private void setDefaultOptions() {
        useIRL = false;
        compileOnly = false;
        compileBasicClasses = false;
        arrayDir = "arraytmp/";
        classDir = "classtmp/";
    }

    public void firstTime() {
       createAuxArrayClasses(); 
    }

    
    private boolean compileClass( 
           String directory, 
           String fileName,
           boolean isRealParamClass
           ) { 
            
       String pathFile = directory + separator + fileName;
       if ( (input = readFile(pathFile)) == null ) 
          return false;
       else {
          MyLexer lexer;
          try {
            lexer = new MyLexer( input, errorMsg );
          } catch ( Exception e ) {
            error.show("Error reading file " + fileName);
            return false;
          }
          errorMsg.setLexer(lexer, fileName);
          
          MyParser parser;

          
          try { 
             parser = new MyParser(lexer, errorMsg, 
               toCreateClassList, toCreateArrayList, newToCreateClassList, 
               isRealParamClass );
          } catch ( Throwable e ) {
            return false;
          }

          GenericClass aClass;
          try {
            if ( fileName.compareTo("ParameterInfo.g") == 0 ) 
               System.out.println("At the chosen file");
            aClass = (GenericClass ) parser.parse();
          } catch ( Throwable e ) {
            System.out.println("Just before printStackTrace");
            e.printStackTrace();
            //error.show("Error compiling " + pathFile);
            return false;
          }
          File outFile;
          try { 
             outFile = new File(pathFile + "i");
          } catch( Exception e ) {
             // error.signal("Error creating file " + pathFile + "i", -1);
             e.printStackTrace();
             return false;
          }
          writeInfoFile( aClass, outFile );
          classList.add(aClass);
          return true;
       }
    }



    private char []readFile( String fileName ) {
        FileReader stream;
        int numChRead;
        
        File file = new File(fileName);
        if ( ! file.exists() || ! file.canRead() ) {
           error.show("Either the file " + fileName + " does not exist or it cannot be read");
           return null;
         }
         try { 
             stream = new FileReader(file);  
         } catch ( FileNotFoundException e ) {
            error.show("Something wrong: file " + fileName + " does not exist anymore");
            return null;
         }
                // one more character for '\0' at the end that will be added by the
                // compiler
         char []input = new char[ (int ) file.length() + 1 ];
            
         try {
            numChRead = stream.read( input, 0, (int ) file.length() );
         } catch ( IOException e ) {
            error.show("Error reading file " + fileName);
            return null;
         }
                
         if ( numChRead != file.length() ) {
            error.show("Read error in file " + fileName);
            return null;
         }
         try {
            stream.close();
         } catch ( IOException e ) {
            error.show("Error in handling the file " + fileName );
            return null;
         }
         return input;
    }
           

    
    public void make() {

          // number of the current line
        lineNumber = 1;
        tokenPos = 0;
        beforeOptions = true;
        inFileName = true;
        nextToken();
        parse();
        Enumeration e;
        if ( ! error.getWasACompilerError() ) {
            
            if ( compileBasicClasses ) {
                // compile the basic classes and the basic array classes 
                compileBasic();
            }
            else {
            
               boolean result = addArrayLib();
               result = addBasicLib() && result;
               result = addWrapperLib() && result;
              
               e = toBeCompiledList.elements();
               while ( e.hasMoreElements() ) {
                  ClassFile cf = (ClassFile ) e.nextElement();
                  result = compileClass( 
                      cf.getDirectory(), cf.getClassName(), false ) && result;
               }
               if ( result ) {
                   createClassesFromClassList();
                   SymbolTable symbolTable = new SymbolTable( classList);
                   e = classList.elements();
                   while ( e.hasMoreElements() ) {
                      GenericClass aClass = classList.get( (String ) e.nextElement() );
                      aClass.reify(classList, errorMsg, symbolTable);
                      symbolTable.deleteClassIdent();
                   }
                   e = classList.elements();
                   while ( e.hasMoreElements() ) {
                      NormalClass aClass = (NormalClass ) classList.get( 
                          (String ) e.nextElement() );
                      try { 
                         aClass.gen( new PrintWriter( new FileOutputStream(      
                             aClass.getJavaName() + ".java") ) ) ;
                      } catch ( Exception except ) {
                         error.show( "I/O error in file " + aClass.getJavaName() + ".java");
                      }
                   }
               }
           }
        }
    }

    
    private void createClassesFromClassList() {
        ToCreateClassList list = toCreateClassList;
        
        while ( list.getSize() != 0 ) {
            Hashtable t = list.getTable();
            Enumeration e = t.keys();
            ClassType classType;
            newToCreateClassList = new ToCreateClassList();
            while ( e.hasMoreElements() ) {
                classType = (ClassType ) t.get( (String ) e.nextElement() );
                createInstantiatedClass(classType);
            }
            e = toCreateArrayList.keys();
            while ( e.hasMoreElements() ) {
                String arrayName = (String ) e.nextElement();
                ArrayType anArray = toCreateArrayList.get(arrayName);
                createArray(anArray);
            }
            list = newToCreateClassList;
        }
    }

    
    void createAnyArray( ArrayType anArray ) {
        
        if ( anArray.getNumDimensions() != 1 ) {
            error.show("Attempt to use an array of non-basic class with " + 
               anArray.getNumDimensions() + " dimensions" );
            return ;
        }
        String dir;
        
        dir = projectDir + separator + arrayDir + separator;
            
        if ( contentsOf_array_any_1_g == null )
           contentsOf_array_any_1_g = CreateArray.readFile( 
                templateDir + separator + "array-any-1.g" );
        // createArrays( array, directory, "$Array$1$_char", "char", "_Char", "char" );

        CreateArray.createArrays( contentsOf_array_any_1_g, dir,
             anArray.getJavaName(), anArray.getType().getName(), null, 
             anArray.getType().getJavaName() );

        if ( contentsOf_iter_array_any_1_g == null )
           contentsOf_iter_array_any_1_g = CreateArray.readFile(
                 templateDir + separator + "iter-any-1.g");

        CreateArray.createArrays( contentsOf_iter_array_any_1_g, dir,
             "$Iter" + anArray.getJavaName(), anArray.getType().getName(), null, 
             anArray.getType().getJavaName() );
                 
    }
    
    private void createArray( ArrayType anArray ) {

        if ( ! (anArray.getType() instanceof AnyValueType ) ) {
            createAnyArray(anArray);
            // it was dir instead of "" - correct !!
            // correct the line below !!!
            // toBeCompiledList( new ClassFile( "", anArray.getJavaName() ) );
        }
    }
    
    
    private void createAuxArrayClasses() {
        
        for ( int i = 0; i < Type.arrayAnyValue.length; i++ ) {
            TypeList typeList = new TypeList();
            typeList.add( new ClassType( Type.arrayAnyValue[i].getJavaName(), 
                null, false ) );
            toCreateClassList.add( new ClassType( "Iter", typeList, false ) );
            toCreateClassList.add( new ClassType( "Function", typeList, false ) );
            toCreateClassList.add( new ClassType( "Command", typeList, false ) );
            toCreateClassList.add( new ClassType( "Filter", typeList, false ) );
        }
    }
            
    
    
    private boolean addBasicLib() {
        
        if ( pathBasicLib == null && ! compileBasicClasses) {
            error.show("No path for basic library classes was specified", -1);
            return false;
        }
        else 
          return addClasses(pathBasicLib);
    }

    private boolean addWrapperLib() {
        
        if ( pathWrapper != null ) 
           return addClasses(pathWrapper);
        else
           return true;
    }

    private boolean addArrayLib() {

        boolean result = true;
        Enumeration e = arrayPathLib.elements();
        while ( e.hasMoreElements() ) 
           if ( ! addClasses( (String ) e.nextElement() ) ) 
               result = false;
        return result;
    }
    
    private boolean addClasses( String path ) {
        
        File file = new File(path);
        if ( ! file.exists() || ! file.canRead() || ! file.isDirectory() ) {
           error.show("The path " + path + 
              " does not exist or I can't read it or it is not a directory", -1);
           return false;
        }
        else {
           boolean ok = true;
           String allFiles[] = file.list();
           Hashtable fileNames = new Hashtable();
           for( int i = 0; i < allFiles.length; i++ ) 
              if ( allFiles[i].endsWith(".g") || allFiles[i].endsWith(".gi") ) 
                  fileNames.put( 
                     allFiles[i].substring(0, allFiles[i].lastIndexOf('.')), "" );
           Enumeration e = fileNames.keys();
           while ( e.hasMoreElements() ) {
              String name = (String ) e.nextElement();
              File fg = new File(path + separator + name + ".g");
              File fgi = new File(path + separator + name + ".gi");
              if ( ! fg.exists() ) {
                    // .g file does not exists
                  if ( ! fgi.exists() ) {
                        // .gi files does not exist
                      error.show("File " + fgi.getPath() + " cannot be read", -1);
                  }
                  else {
                     GenericClass aClass = readInfoFile(fgi);
                     if ( aClass == null ) {
                        ok = false;
                     }
                     else
                       classList.add(aClass);
                  }
              }
              else 
                  if ( ! fgi.exists() || fg.lastModified() > fgi.lastModified() ) {
                      //error.show( "File "  + fg.getPath() + " needs to be compiled or it does not exist", -1);
                      //ok = false;
                      ok = compileClass(path, name + ".g", false);
                  }
                  else {
                     GenericClass aClass = readInfoFile(fgi);
                     if ( aClass == null ) {
                        ok = false;
                     }
                     else
                       classList.add(aClass);
                  }
           }
          return true;
       }
    }
    
    
    private GenericClass readInfoFile( File fgi ) {
        
        int k;
        FileInputStream istream;
        ObjectInputStream p;
        GenericClass aClass;
        String name;
        try {
           istream = new FileInputStream(fgi);
	       p = new ObjectInputStream(istream);
	       try { 
              aClass = (GenericClass ) p.readObject();
           } catch ( Exception e ) {
              error.show("Probably file " + fgi.getPath() +  " is from a outdated format of GenericClass");
              return null;
           }
           name = fgi.getName();
           if ( (k = name.lastIndexOf('.')) >= 0 ) 
              name = name.substring(0, k);
           if ( aClass.getName().compareTo(name) != 0 )  {
               error.show("File " + fgi.getPath() + " has a class with name " + aClass.getName(), -1 );
               return null;
           }
           else
              return aClass;           
        } catch ( Exception e ) {
            error.show("Error reading file " + fgi.getPath() );
            return null;
        }
    }


    private void compileBasic() {
        
        File file = new File(pathBasicLib);
        
        String allFiles[] = file.list();
        for( int i = 0; i < allFiles.length; i++ ) 
           if ( allFiles[i].endsWith(".g") ) 
              if ( ! compileClass(pathBasicLib, allFiles[i], false) ) {
                 error.signal("Fatal error while compiling a basic class", -1);
                 return;
              }
    }    

    private void createInstantiatedClass( ClassType classType ) 
       /* create a real class from classType, which is something like List(char) or Stack(Person)
          Create .g and .gi files, put in projectDir/classDir */
    {
        NormalClass aClass = (NormalClass ) classList.get( classType.getName() );
        if ( aClass == null ) {
            error.show("Parameterized class " + classType.getName() + " not found");
            return;
        }
        
        if ( aClass.getParamList() == null ) {
            error.show("Class " + aClass.getName() + " used as a parameterized class");
            return;
        }
        
        Hashtable paramTable = new Hashtable(); //  ( "T", "{ char, null }" }
        ClassParameter formalParameter;
        ClassType realParameter;
        Enumeration formalList = aClass.getParamList().elements();
        Enumeration realList = classType.getTypeList().elements();
        if ( aClass.getParamList().size() != classType.getTypeList().getSize() ) {
           error.show("Parameterized class " + aClass.getName() + " used with wrong number of parameters");
           return;
        }
        while ( formalList.hasMoreElements() ) {
            formalParameter = (ClassParameter ) formalList.nextElement();
            realParameter = (ClassType ) realList.nextElement();
            paramTable.put( formalParameter.getName(), realParameter.getJavaName() );
            // T <->  char in List(char) or
            // T <-> $pc$List$char in Queue(List(char))
            // result:  $pc$Queue$1$pc$List$char
        }
        String pcFileName;
        char []inputParamClass;
        pcFileName = aClass.getJavaName() + ".g";
        try {
            inputParamClass = readFile( pcFileName );
        } catch ( Exception e ) {
            error.show("Error reading file " + pcFileName);
            return;
        }
        if ( inputParamClass == null ) 
           return;
        PrintWriter out = null;
        try { 
            out = new PrintWriter( new FileOutputStream( 
               projectDir + separator + classDir + separator + classType.getJavaName() + ".g" ) );
        } catch ( Exception e ) {
            error.show("Error opening file ");
            return;
        }
        ParamClassLexer pcLexer = new ParamClassLexer( inputParamClass, 
            errorMsg, paramTable,  out );
        
        pcLexer.makeRealClass();
        
           // see the use of the code below
        File realClassFile = new File( classDir + classType.getJavaName() + ".g");
        FileOutputStream fout;
        try {
            fout = new FileOutputStream(realClassFile);
        } catch ( Exception e ) {
            error.show("Error in handling file of parameterized class "  +
                classType.getJavaName() );
            return ;
        }
        
        // call the lexer ...
        String directory = projectDir + separator + classDir;
        toBeCompiledList.addElement( new ClassFile(directory, classType.getJavaName()));
        if ( ! compileClass(directory, classType.getJavaName(), false) ) 
           error.show("Error in compilation ...");
           
    }
          

    private boolean writeInfoFile( GenericClass aClass, File fgi ) {
        
        try { 
           FileOutputStream ostream = new FileOutputStream(fgi);
           ObjectOutputStream p = new ObjectOutputStream(ostream);
           p.writeObject(aClass);
           ostream.close();
        } catch( Exception e ) {
            error.show("Can't write to file " + fgi.getPath() );
            return false;
        }
        return true;
    }

    private void parse() {
        fileNameList();
        optionsList();
        if ( token.tk != PSymbol.EOF )
          error.show("EOF expected", lineNumber);
    }
    
    private void fileNameList() {
        
        while ( token.tk == PSymbol.STRING || token.tk == PSymbol.PATHNAME ) {
            String name  = (String ) token.getValue();
            if ( ! name.endsWith(".g") ) 
                name = name + ".g";
            toBeCompiledList.addElement(new ClassFile(projectDir, name) );
            nextToken();
        }
    }
            
    private void optionsList() {
        if ( token.tk != PSymbol.OPTIONS ) 
          error.show("Options expected", lineNumber);
        else {
            inFileName = false;
            nextToken();
            while ( token.tk != PSymbol.EOF )
              option();
        }
    }

    
    private void option() {
        switch ( token.tk ) {
            case PSymbol.BASICLIB :
              nextToken();
              if ( token.tk != PSymbol.ASSIGN )
                error.show("= expected", lineNumber);
              else {
                inFileName = true;
                nextToken();
                inFileName = false;
                if ( token.tk != PSymbol.STRING && token.tk != PSymbol.PATHNAME )
                  error.show("File name expected", lineNumber);
                else {
                    pathBasicLib = (String ) token.getValue();
                    nextToken();
                }
              }
              break;
            case PSymbol.LIB :
              nextToken();
              if ( token.tk != PSymbol.ASSIGN )
                error.show("= expected", lineNumber);
              else {
                inFileName = true;
                nextToken();
                inFileName = false;
                if ( token.tk != PSymbol.STRING && token.tk != PSymbol.PATHNAME )
                  error.show("File name expected", lineNumber);
                else {
                    arrayPathLib.addElement( (String ) token.getValue() );
                    nextToken();
                }
              }
              break;
            case PSymbol.JAVALIB :
              nextToken();
              if ( token.tk != PSymbol.ASSIGN )
                error.show("= expected", lineNumber);
              else {
                inFileName = true;
                nextToken();
                inFileName = false;
                if ( token.tk != PSymbol.STRING && token.tk != PSymbol.PATHNAME )
                  error.show("File name expected", lineNumber);
                else {
                    arrayJavaLib.addElement( (String ) token.getValue() );
                    nextToken();
                }
              }
              break;
            case PSymbol.WRAPPER:
              nextToken();
              if ( token.tk != PSymbol.ASSIGN )
                error.show("= expected", lineNumber);
              else {
                inFileName = true;
                nextToken();
                inFileName = false;
                if ( token.tk != PSymbol.STRING && token.tk != PSymbol.PATHNAME )
                  error.show("File name expected", lineNumber);
                else {
                    pathWrapper = (String ) token.getValue();
                    nextToken();
                }
              }
              break;
            case PSymbol.TEMPLATEDIR : 
              nextToken();
              if ( token.tk != PSymbol.ASSIGN )
                error.show("= expected", lineNumber);
              else {
                inFileName = true;
                nextToken();
                inFileName = false;
                if ( token.tk != PSymbol.STRING && token.tk != PSymbol.PATHNAME )
                  error.show("File name expected", lineNumber);
                else {
                    templateDir = (String ) token.getValue();
                    /* if ( templateDir.length > 0 && 
                         templateDir.charAt(templateDir.length - 1) != separator.charAt(0) )
                       templateDir = templateDir + separator; */
                    nextToken();
                }
              }
              break;
            case PSymbol.PROJECTDIR: 
              nextToken();
              if ( token.tk != PSymbol.ASSIGN )
                error.show("= expected", lineNumber);
              else {
                inFileName = true;
                nextToken();
                inFileName = false;
                if ( token.tk != PSymbol.STRING && token.tk != PSymbol.PATHNAME )
                  error.show("File name expected", lineNumber);
                else {
                    projectDir = (String ) token.getValue();
                    nextToken();
                    Enumeration e = toBeCompiledList.elements();
                    while ( e.hasMoreElements() ) {
                       ClassFile cf = (ClassFile ) e.nextElement();
                       cf.setDirectory(projectDir);
                    }
                }
              }
              break;
            case PSymbol.USEIRL :
              useIRL = true;
              nextToken();
              break;
            case PSymbol.COMPILEONLY :
              compileOnly = true;
              nextToken();
              break;
            case PSymbol.COMPILEBASICCLASSES :
              compileBasicClasses = true;
              nextToken();
              break;
            case PSymbol.ARRAYDIR :
              nextToken();
              if ( token.tk != PSymbol.ASSIGN )
                error.show("= expected", lineNumber);
              else {
                inFileName = true;
                nextToken();
                inFileName = false;
                if ( token.tk != PSymbol.STRING && token.tk != PSymbol.PATHNAME )
                  error.show("Directory name expected", lineNumber);
                else {
                    arrayDir = (String ) token.getValue();
                    nextToken();
                }
              }
              break;
            case PSymbol.CLASSDIR:
              nextToken();
              if ( token.tk != PSymbol.ASSIGN )
                error.show("= expected", lineNumber);
              else {
                inFileName = true;
                nextToken();
                inFileName = false;
                if ( token.tk != PSymbol.STRING && token.tk != PSymbol.PATHNAME )
                  error.show("Directory name expected", lineNumber);
                else {
                    classDir = (String ) token.getValue();
                    nextToken();
                }
              }
              break;
            default:
              error.show("Option expected", lineNumber);
              
        }
    }
              
        
        
        
    
    private void nextToken() {
        char ch;
        
        while (  (ch = input[tokenPos]) == ' ' || ch == '\r' ||
                 ch == '\t' || ch == '\n')  {
            // count the number of lines
          if ( ch == '\n')
            lineNumber++;
          tokenPos++;
          }
        if ( ch == '\0') 
          token = new PSymbol(PSymbol.EOF);
        else
          if ( input[tokenPos] == '/' && input[tokenPos + 1] == '/' ) {
                // comment found
               while ( input[tokenPos] != '\0'&& input[tokenPos] != '\n' )
                 tokenPos++;
               nextToken();
               }
          else {
            if ( inFileName ) {
                if ( ch == '\"' ) {
                    tokenPos++;
                    getString();
                }
                else {
                    StringBuffer ident = new StringBuffer();
                    while ( ! Character.isWhitespace(input[tokenPos]) 
                            && input[tokenPos] != '\0' ) {
                        ident.append( input[tokenPos] );
                        tokenPos++;
                    }
                    String s = ident.toString();
                    if ( beforeOptions && s.toLowerCase().compareTo("options") == 0 ) {
                        beforeOptions = false;
                        token = new PSymbol( PSymbol.OPTIONS );
                    }
                    else
                        token = new PSymbol( PSymbol.PATHNAME, s );
                    }
            }
            else {

               if ( Character.isLetter( ch ) ) {
                   // get a keyword
                   StringBuffer ident = new StringBuffer();
                   while ( Character.isLetter( input[tokenPos] ) ) {
                       ident.append(input[tokenPos]);
                       tokenPos++;
                   }
                   String identStr = ident.toString().toLowerCase();
                      // if identStr is in the list of keywords, it is a keyword !
                   Object value = keywordsTable.get(identStr);
                   if ( value == null ) 
                     error.show( "Option " + ident + " is not valid", lineNumber );
                   else 
                       // get the token
                     token = (PSymbol ) value;
                   if ( Character.isDigit(input[tokenPos]) )
                     error.show("Word followed by a number", lineNumber);
               }
               else if ( Character.isDigit( ch ) ) {
                   // get a number
                   StringBuffer number = new StringBuffer();
                   while ( Character.isDigit( input[tokenPos] ) ) {
                       number.append(input[tokenPos]);
                       tokenPos++;
                   }
                   token =  new PSymbol( PSymbol.NUMBER, new Integer(number.toString()) );
               } else {
                   tokenPos++;
                   switch ( ch ) {
                       case '=' :
                         token = new PSymbol( PSymbol.ASSIGN );
                         break;
                       case '"' :
                          getString();
                          break;
                       default :
                         error.show("Invalid Character: '" + ch + "'", lineNumber);
                       }
                   }
               }
             }
    }

     
    private void getString() {
        
        StringBuffer ident = new StringBuffer();
        while ( input[tokenPos] != '\"' && input[tokenPos] != '\0' ) {
           ident.append(input[tokenPos]);
           tokenPos++;
        }
        String identStr = ident.toString();
        token = new PSymbol( PSymbol.STRING, identStr );
        if ( input[tokenPos] == '\0' ) 
           error.signal("End-of-file unexpected", lineNumber);
        else
           tokenPos++;
    }
    

      // contains the keywords
    static private Hashtable keywordsTable;

     // this code will be executed only once for each program execution
     static {
        keywordsTable = new Hashtable();
        keywordsTable.put( "options", new PSymbol(PSymbol.OPTIONS) );
        keywordsTable.put( "basiclib", new PSymbol(PSymbol.BASICLIB) );
        keywordsTable.put( "lib", new PSymbol(PSymbol.LIB) );
        keywordsTable.put( "wrapper", new PSymbol(PSymbol.WRAPPER) );
        keywordsTable.put( "useirl", new PSymbol(PSymbol.USEIRL) );
        keywordsTable.put( "compileonly", new PSymbol(PSymbol.COMPILEONLY) );
        keywordsTable.put( "compilebasicclasses", new PSymbol(PSymbol.COMPILEBASICCLASSES) );
        keywordsTable.put( "arraydir", new PSymbol(PSymbol.ARRAYDIR) );
        keywordsTable.put( "classdir", new PSymbol(PSymbol.CLASSDIR) );
        keywordsTable.put( "templatedir", new PSymbol(PSymbol.TEMPLATEDIR) );
        keywordsTable.put( "projectdir", new PSymbol(PSymbol.PROJECTDIR) );
     }
     
      // current token
    private PSymbol token;
    private int  tokenPos, lineNumber;
      // program given as input - source code
    private char []input;
    private String name, projectDir, separator;
    private ProjectError error;
    private ErrorMsg errorMsg;
    private SymbolTable symbolTable;

    private String pathBasicLib; // basic classes and IRL
    private String pathWrapper;  // Wrapper classes
    
    private boolean useIRL, compileOnly, compileBasicClasses;
    
    private int state;
    
    private Vector arrayPathLib, arrayJavaLib, toBeCompiledList;
    private ClassList classList;
    private String arrayDir, classDir, templateDir;
    char []contentsOf_array_any_1_g;
    char []contentsOf_iter_array_any_1_g;
    private ToCreateClassList toCreateClassList, newToCreateClassList;
    private ToCreateArrayList toCreateArrayList;
    private boolean beforeOptions, inFileName;
}
    
